#!/usr/local/bin/wish -f
#	Installation Notes (see also the README file):
# 1. Change the first line above to point to the wish on your system
# 2. If you use cc instead of gcc, update the CC line below
	set Unix 1      ;# on PC: set to 0 (Windows95 or WindowsNT)
	set CC   "gcc"  ;# on PC: use gcc
	set GO32 ""     ;# on PC: use go32 (also with v2 gcc)
	set CMD  "command.com";# for WindowsNT PC use: "cmd" instead
	set SPIN "spin" ;# or use a full path-name if necessary
	set BG   "white";# default background color for text
	set FG   "black";# default foreground color for text

#=====================================================================#
# Tcl/Tk Spin Controller Version 3.0.0, 19 July 1997                   #
# written by Gerard J. Holzmann, Bell Laboratories, USA.              #
# Copyright (c) 1995-1997 by Lucent Technologies. All Rights Reserved.#
# See the README file for more detailed installation notes.           #
#=====================================================================#

set xversion "3.0.0 -- 19 July 1997"

wm title . "SPIN CONTROL $xversion"
wm iconname . "SPIN"
wm geometry . +100+100
wm minsize  . 400 200

set Fname ""

# allow no more than one entry per selection
catch { tk_listboxSingleSelect Listbox }

proc msg_file {f nowarn} {
	set msg ""
	set ef [open $f r]
	while {[gets $ef line] > -1} {
		if {$nowarn} {
			if {[string first "warning" $line] < 0} {
				set msg "$msg$line"
			}
		} else {
			set msg "$msg$line"
	}	}
	close $ef
	return $msg
}

	scan $tk_version "%d.%d" tk_major tk_minor
	if {$Unix} {
		set version [exec $SPIN -V]
	} else {
		catch { exec $SPIN -V >&pan.tmp } err
		if {$err == ""} {
			set version [msg_file pan.tmp 1]
		} else {
			puts "error: $err"
			puts "is there a $SPIN and a go32.exe?"
			exit
		}
	}
	# warn if installed spin is too old
	if {[string first "Spin Version 2" $version] < 0 \
	&&  [string first "Spin Version 3" $version]} {
		set version "Installed Spin: older than Version 2.0"
	}

frame .menu
	# top level menu bar
	menubutton .menu.file -text "File.." -relief raised -menu .menu.file.m
	button .menu.ltl -text "LTL.." -command call_tl
	button .menu.val -text "Verify.." -fg red \
		-command validation
	button .menu.sim -text "Simulate.." -fg blue \
		-command simulation
	button .menu.fsm -text "Fsm view.." -command fsmview
	button .menu.stc -text "Check Syntax" -fg darkgreen \
		-command syntax_check

	label .menu.lno -text "Line#:" -relief raised 
	entry .menu.ent -width 6 -textvariable lno -relief sunken
	bind  .menu.ent <Return> {
		.inp.t tag remove hilite 0.0 end
		.inp.t tag add hilite $lno.0 $lno.1000
		.inp.t tag configure hilite -background $FG -foreground $BG
		.inp.t yview -pickplace [expr $lno-1]
	}

	label .menu.fnd1 -text "Find:" -relief raised 
	entry .menu.fnd2 -width 8 -textvariable pat -relief sunken 
	bind  .menu.fnd2 <Return> {
		.inp.t tag remove hilite 0.0 end
		forAllMatches .inp.t $pat
	}

	menubutton .menu.help -text "Help" -relief raised \
		-menu .menu.help.m

	pack append .menu \
		.menu.file {left frame w} \
		.menu.help {left frame w} \
		.menu.stc {left frame w} \
		.menu.sim {left frame w} \
		.menu.val {left frame w} \
		.menu.ltl {left frame w} \
		.menu.fsm {left frame w} \
		.menu.lno {left frame w} \
		.menu.ent {left frame w} \
		.menu.fnd1 {left frame w} \
		.menu.fnd2 {left frame w}

frame .inp
	# view of spin input
	scrollbar .inp.s \
		-command ".inp.t yview"
	text      .inp.t -bd 2 \
		-yscrollcommand ".inp.s set" \
		-wrap word
	pack append .inp \
		.inp.s {left filly} \
		.inp.t {left expand fill}

set l_typ 0;	# used by both simulator and validator
set ol_typ -1;	# remembers setting last used in compilation
set m_typ 2;	# used by simulator

menu .menu.file.m
	.menu.file.m add command -label "Clear"  -command ".inp.t delete 0.0 end"
	.menu.file.m add command -label "UnSelect" \
		-command ".inp.t tag remove hilite 0.0 end;\
			  .inp.t tag remove sel 0.0 end"
	.menu.file.m add command -label "Open.." -command "browse"
	.menu.file.m add command -label "ReOpen" -command "reopen"
	.menu.file.m add command -label "Save.." -command "save_spec"
	.menu.file.m add command -label "Quit" \
		-command "cleanup; destroy .; exit"

menu .menu.help.m
	.menu.help.m add command -label "Xspin Usage" \
		-command "helper"
	.menu.help.m add command -label "Simulation" \
		-command "roadmap1"
	.menu.help.m add command -label "Verification" \
		-command "roadmap2"
	.menu.help.m add command -label "LTL Formulae" \
		-command "roadmap4"
	.menu.help.m add command -label "FSM Views" \
		-command "roadmap5"
	.menu.help.m add command -label "Reducing Complexity" \
		-command "roadmap3"

set Log ""
set logwindow 0
set maxlog 0

proc cleanup {} {
	global Unix
	if {$Unix == 0} {
		add_log "removing tmp files.."; update
	}
	rmfile "pan.h pan.c pan.t pan.m pan.b pan.tmp"
	rmfile "pan.oin pan.pre pan.out pan.exe"
	rmfile "pan_in pan_in.trail trail.out pan"
	if {$Unix == 0} {
		add_log "done.."
	}
}

proc show_log {} {
	global Log

	catch { destroy .out };	# probably redundant
	toplevel .out
	wm title .out "SPIN LOG"
	wm iconname .out "SpinLog"

	text .out.t -bd 2 -setgrid 1 -width 30 -height 15 \
		-font -Adobe-Helvetica-Bold-R-Normal-*-120-* 
	pack append .out .out.t {left expand fill}
	.out.t insert end $Log
}

proc add_log {{y ""}} {
	global Log maxlog logwindow

	if {$logwindow == 0} {
		puts "xspin:  $y"
		return
	}

	set Log "${Log}$y\n"
	catch {
		.out.t insert end "$y"
		.out.t yview -pickplace end
		.out.t insert end "\n"
		if {[string length $y] > $maxlog} {
			set maxlog [string length $y]
			.out.t configure -width $maxlog
		}
	}
}

	if {$logwindow} {
		show_log
		# wait for the spin log window to exist
		tkwait visibility .out
	}


pack append . \
	.inp  {bot frame w expand fill} \
	.menu {top fillx}

# simulation parameters - initial settings
	set fvars 1
	set gvars 1;	set lvars 1
	set msc   1;	set svars 1
	set rvars 1;	set s_typ 0
	set stop  0;	set tsc 1
	set seed	"1";	# random sumulation
	set jumpsteps	"0";	# guided simulation

add_log " $version"
add_log "Xspin Version $xversion"
add_log "TclTk Version [info tclversion]/$tk_version "

.inp.t configure -background $BG -foreground $FG

# xspin
# process execution barchart

set Data(0) 0
set Name(0) "-"
set n 0
set bar_handle 0

proc stopbar {} {
	global Data Name n
	for {set i 0} {$i <= $n} {incr i} {
		set Data($i) 0
		set Name($i) ""
	}
	set n 0
	catch { destroy .bar }
}

proc growbar {v} {
	global n Data
	set Data($n) $v
	incr n
	catch { fillbar }
}

proc shrinkbar {} {
	global n
	incr n -1
	set Data($n) 0
	catch { fillbar }
}

proc stepbar {v nm} {
	global n Data Name

	if {$v >= 0} {
		if { [info exists Data($v)] } {
			incr Data($v)
		} else {
			set Data($v) 1
		}
		if {$v >= $n} {
			set n [expr $v+1]
		}
		if { [string length $nm] > 4} {
			set Name($v) [string range $nm 0 4]
		} else {
			set Name($v) $nm
		}
		catch { fillbar }
	}
}

proc adjustbar {v w} {
	global Data

	set Data($v) $w
	catch { fillbar }
}

proc startbar {tl} {
	global n Data bar_handle

	catch { destroy .bar }
	toplevel .bar
	wm minsize .bar 200 200
	wm title .bar "$tl"

	set maxy [expr [winfo screenheight .] - 200]

	wm geometry .bar +[expr [winfo rootx .]+150]+$maxy

	set bar_handle [canvas .bar.c -height 410 -width 410 -relief raised]
	frame  .bar.buts

	button .bar.buts.s1 -text "Save in:  panbar.bw.ps" \
		-command ".bar.c postscript -file panbar.bw.ps -colormode mono"
	button .bar.buts.s2 -text "Save in:  panbar.col.ps" \
		-command ".bar.c postscript -file panbar.col.ps -colormode color"

	button .bar.buts.b -text " Close " -command "stopbar"
	pack .bar.buts.b .bar.buts.s1 .bar.buts.s2 -side right
	pack append .bar  .bar.c {top expand fill}  .bar.buts {bot frame e}
}

proc fillbar {} {
	global n Data Name

	.bar.c delete grid
	.bar.c delete data
	set sum 0
	for {set i 0} {$i < $n} {incr i} {
		if { [info exists Data($i)] } {
			incr sum $Data($i)
		} else {
			set Data($i) 0
			set Name($i) "-"
		}
	}
	for {set i 0} {$i < $n} {incr i} {
		.bar.c create line \
			$i 0 \
			$i 100 \
			-fill #222222 -tags grid
		.bar.c create text $i 105 \
			-text $i -tags grid
		.bar.c create text $i 110 \
			-text "$Name($i)" \
			-fill blue -tags grid

		if { [info exists Data($i)] } {
			set y [expr ($Data($i)*100)/$sum]
			.bar.c create line \
				$i 100  \
				$i [expr 100-$y] \
				-width 35 -fill red -tags data
			if {$y > 6} {
				set nrcol "yellow"
			} else {
				set nrcol "red"
			}
			.bar.c create text $i 95 \
				-text "$Data($i)" \
				-fill $nrcol -tags grid
		}
	}

	.bar.c create text [expr ($n)/2.0] -15 -text "Percentage of $sum System Steps" \
		-anchor c -justify center -tags grid
	.bar.c create text [expr ($n)/2.0] -8 -text "Executed Per Process ($n total)" \
		-anchor c -justify center -tags grid
	.bar.c scale all 0 0 55 3
	if {$n <= 5} {
		.bar.c move all 100 60
	} else {
		.bar.c move all 50 60
	}
}

proc barscales {} {
	global bar_handle

	catch {
		button .bar.buts.b4 -text "Larger" \
			-command "$bar_handle scale all 0 0 1.1 1.1"
		button .bar.buts.b5 -text "Smaller" \
			-command "$bar_handle scale all 0 0 0.9 0.9"
		pack append .bar.buts \
			.bar.buts.b4 {right padx 5} \
			.bar.buts.b5 {right padx 5}
	}
}
# xspin
# Files and Generic Boxes

set file ""
set boxnr 0

proc rmfile {f} {
	global Unix CMD tk_major tk_minor
	if {$Unix} {
	#	add_log "rm $f"
		catch {eval exec rm -f $f}
	} else {
		set n [llength $f]
		for {set i 0} {$i < $n} {incr i} {
			set g [lindex $f $i]
			add_log "rm $g"
			if {$tk_major >= 4 && $tk_minor >= 2} {
				catch {eval exec $CMD /c del $g}
			} else {
				catch {eval exec $CMD >&@stdout /c del $g}
			}
	}	}
}

proc mvfile {f g} {
	global Unix CMD tk_major tk_minor
	if {$Unix} {
		catch {exec mv $f $g}
	} else {
		if {$tk_major >= 4 && $tk_minor >= 2} {
			catch {eval exec $CMD /c move $f $g}
		} else {
			catch {eval exec $CMD >&@stdout /c move $f $g}
		}
	}
}

proc cpfile {f g} {
	global Unix CMD tk_major tk_minor
	if {$Unix} {
		catch {exec cp $f $g}
	} else {
		if {$tk_major >= 4 && $tk_minor >= 2} {
		catch {eval exec $CMD /c copy $f $g}
		} else {
		catch {eval exec $CMD >&@stdout /c copy $f $g}
	}	}
}

proc cmpfile {f g} {
	global Unix

	set err ""
	if {$Unix} {
		catch {exec cmp $f $g} err
	} else {
		if {[file exists $f] == 0 \
		||  [file exists $g] == 0} {
			return "error"
		}
		set fd1 [open $f r]
		set fd2 [open $g r]
		while {1} {
			set n1 [gets $fd1 line1]
			set n2 [gets $fd2 line2]
			if {$n1 != $n2 \
			||  [string compare $line1 $line2] != 0} {
				set err "files differ"
				break
			}
			if {$n1 < 0} { break }
		}
		close $fd1
		close $fd2
	}
	return $err
}

proc file_ok {f} {
	if {[file exists $f]} {
		if {![file isfile $f] || ![file writable $f]} {
			add_log "error: file $f is not writable"
			return 0
	}	}
	return 1
}

proc mkpan_in {} {
	set fd [open pan_in w]
	puts $fd [.inp.t get 0.0 end] nonewline
	catch "flush $fd"
	catch "close $fd"
}

proc no_change {} {

	mkpan_in
	if {![file exists pan.oin]} {
		cpfile pan_in pan.oin
		if {![file exists pan.oin]} {
			add_log "err: cannot copy files"
			return 2	; bad news
		}
		return 0		; first time
	}
	set err [cmpfile pan_in pan.oin]
	if {[string length $err] > 0} {
		cpfile pan_in pan.oin
		return 0		;# different
	}
	return 1			;# unchanged
}

proc mk_exec {} {
	global Unix CC SPIN

	set nochange [no_change]
	if {$nochange == 2} { return 2 }
	if {$nochange == 1 && [file exists "pan"]} {
		add_log "<no recompilation needed>"
		return 1
	}

	add_log "<compiling executable>"
	catch { warner }
	add_log "$SPIN -a pan_in"

	catch { exec $SPIN -a pan_in >&pan.tmp }
	set errmsg [msg_file pan.tmp 1]

	if {[string length $errmsg]>0} {
		add_log "$errmsg"
		add_log "<stopped compilation attempt>"
		catch { destroy .warn }
		return 0
	}

	add_log "$CC -o pan -D_POSIX_SOURCE pan.c"; update
	if {$Unix} {
		catch { eval exec $CC -o pan -D_POSIX_SOURCE pan.c } errmsg
	} else {
		catch { eval exec $CC -o pan -D_POSIX_SOURCE pan.c >&pan.tmp }
		set errmsg [msg_file pan.tmp 0]
	}
		if {[string length $errmsg]>0} {
			add_log "$errmsg"
			catch { destroy .warn }
			return 0
		}
	add_log "<compilation complete>"
	catch {destroy .warn}
	return 1
}

proc warner {} {

	catch {destroy .warn}
	toplevel .warn

	wm title .warn "Update"
	wm iconname .warn "Update"
	wm geometry .warn +200+200

	message .warn.t -width 300 -text "\
Please wait until compilation of the \
executable produced by spin completes."

	pack append .warn .warn.t {top expand fill}
	update
}

proc dosave {} {
	global Fname xversion

	if {[file_ok $Fname]==0} return
	set fd [open $Fname w]
	add_log "<saved spec in $Fname>"
	puts $fd "[.inp.t get 0.0 end]" nonewline
	catch "flush $fd"
	catch "close $fd"
	wm title . "SPIN CONTROL $xversion -- File: $Fname"
}

proc save_spec {} {
	global Fname

	catch {destroy .prompt2}
	toplevel .prompt2

	wm title .prompt2 "Save"
	wm iconname .prompt2 "Save"
	wm geometry .prompt2 +200+200

	frame .prompt2.top -relief flat
	label .prompt2.top.lb -text "File name:"
	entry .prompt2.top.fn \
		-textvariable Fname \
		-width 25 \
		-relief raised
	pack append .prompt2.top \
		.prompt2.top.lb {left frame w} \
		.prompt2.top.fn {right expand fillx}

	button .prompt2.ok -text "Ok" \
		-command "destroy .prompt2; dosave"
	button .prompt2.nk -text "Cancel" \
		-command "destroy .prompt2"

	pack append .prompt2 \
		.prompt2.top {top frame w fillx} \
		.prompt2.ok {right} \
		.prompt2.nk {right}
}

proc browse {} {
	global file Fname xversion

	catch {destroy .b}
	toplevel .b

	wm iconname .b "Open"

	scrollbar .b.scroll -command ".b.list yview"
	listbox .b.list -yscroll ".b.scroll set" -relief raised 

	pack append .b \
		.b.scroll {right filly} \
		.b.list {left expand fill}

	fillerup
	bind .b.list <Double-Button-1> {
		set file [selection get]
		if {[file isdirectory $file]} then {
			cd $file
			fillerup
			add_log "cd $file"
		} else {
			if {![file isfile $file]} {
				set file ""
			} else {
				readinfile .inp.t $file
				set dir [pwd]
				set Fname $file
				wm title . "SPIN CONTROL $xversion -- File: $Fname"
				destroy .b
			}
		}
	}
}

proc browse_tl {} {
	catch {destroy .b}
	toplevel .b
	wm iconname .b "Load"

	scrollbar .b.scroll -command ".b.list yview"
	listbox .b.list -yscroll ".b.scroll set" -relief raised 

	pack append .b \
		.b.scroll {right filly} \
		.b.list {left expand fill}

	fillerup
	bind .b.list <Double-Button-1> {
		set file2 [selection get]
		if {[file isdirectory $file2]} then {
			cd $file2
			fillerup
			add_log "cd $file2"
		} else {
			if {![file isfile $file2]} {
				set file2 ""
			} else {
				readinfile .tl.never.t $file2
				set dir [pwd]
				destroy .b
			}
		}
	}
}

proc reopen {} {
	global Fname

	catch {readinfile .inp.t $Fname} ermsg
	if {[string length $ermsg]<=1} return
	add_log $ermsg
}

proc readinfile {into from} {
	global LastGenerate

	set maxlngth 0

	$into delete 0.0 end
	set fd [open $from r]
	
	while {[gets $fd line] > -1} {
		$into insert end "$line\n"
		if {[string length $line] > $maxlngth} {
			set maxlngth [string length $line]
		}
	}
	catch "close $fd"

	if {$maxlngth > 80} { set maxlngth 80 }
#	this resizing is kind of annoying, so disabled:
#	$into configure -width $maxlngth

	add_log "<open $from>"
	set LastGenerate ""	;# used in Validation.tcl
}

proc fillerup {} {
	wm title .b [pwd]
	.b.list delete 0 end

	set dotdot 0
	set l [lsort [glob *]]
	foreach i $l {
		.b.list insert end $i
		if {$i == ".."} { set dotdot 1 }
	}
	if {$dotdot == 0} { .b.list insert 0 ".." }
}

proc gotoline {} {
	global BG FG

	catch { destroy .ln }
	toplevel .ln
	wm title .ln "Goto Line"
	wm iconname .ln "Goto"

	label .ln.lab -text "Enter line number:"
	entry .ln.ent -width 20 -relief sunken -textvariable lno
	pack append .ln \
		.ln.lab {left padx 1m} \
		.ln.ent {right expand}
	bind .ln.ent <Return> {
		.inp.t tag remove hilite 0.0 end
		.inp.t tag add hilite $lno.0 $lno.1000
		.inp.t tag configure hilite \
			-background $FG -foreground $BG
		.inp.t yview -pickplace [expr $lno-1]
	}
	focus .ln.ent
}

proc savebox {b} {
	set fname [.c$b.f.e1 get]
	if {[file_ok $fname]==0} return
	set fd [open $fname w]
	add_log "<saved output in $fname>"
	puts $fd "[.c$b.z.t get 0.0 end]" nonewline
	catch "flush $fd"
	catch "close $fd"
}

proc mkbox {n m p {wd 60} {ht 10} {xp 200} {yp 200}} {
	global boxnr FG BG

	incr boxnr

	toplevel .c$boxnr
	wm title .c$boxnr $n
	wm iconname .c$boxnr $m
	wm geometry .c$boxnr +$xp+$yp

	frame .c$boxnr.z
	text .c$boxnr.z.t -relief raised -bd 2 \
		-background $BG -foreground $FG \
		-yscrollcommand ".c$boxnr.z.s set" \
		-setgrid 1 -width $wd -height $ht -wrap word
	scrollbar .c$boxnr.z.s \
		-command ".c$boxnr.z.t yview"
	pack append .c$boxnr.z \
		.c$boxnr.z.s {left filly} \
		.c$boxnr.z.t {left expand fill}

	button .c$boxnr.b -text "Close" \
		-command "destroy .c$boxnr"
	button .c$boxnr.c -text "Clear" \
		-command ".c$boxnr.z.t delete 0.0 end"
	pack append .c$boxnr \
		.c$boxnr.z {top expand fill} \
		.c$boxnr.b {right padx 5} \
		.c$boxnr.c {right padx 5}

	if {[string length $p]>0} {
		frame  .c$boxnr.f -relief sunken
		button .c$boxnr.f.e0 -text "Save in: " \
			-command "savebox $boxnr"
		entry  .c$boxnr.f.e1 -relief flat -width 10
		.c$boxnr.f.e1 insert end $p
		pack append .c$boxnr.f \
			.c$boxnr.f.e0 {left padx 5} \
			.c$boxnr.f.e1 {left}
		pack append .c$boxnr \
			.c$boxnr.f {right padx 5}
	}

	tkwait visibility .c$boxnr
	return $boxnr
}

# Tcl/Tk book, page 219
proc forAllMatches {w pattern} {
	global BG FG

	scan [$w index end] %d numLines
	for {set i 1} {$i < $numLines} { incr i} {
		$w mark set last $i.0
		if {[regexp -indices $pattern \
			[$w get last "last lineend"] indices]} {
				$w mark set first \
					"last + [lindex $indices 0] chars"
				$w mark set last "last + 1 chars \
					+ [lindex $indices 1] chars"
			.inp.t tag add hilite $i.0 $i.1000
			.inp.t tag configure hilite \
				-background $FG -foreground $BG
			.inp.t yview -pickplace [expr $i-1]
		}
	}
}

set noprep 0
set hasfile 0

proc hasWord {pattern} {
	global SPIN Fname noprep hasfile

	set errmsg ""
	if {$hasfile == 0 && $noprep == 0} {
		add_log "$SPIN -Z pan_in ;# preprocess input"
		catch { eval exec $SPIN -Z pan_in >&pan.tmp } errmsg
		# leaves output in pan.pre
		add_log "<done preprocess>"
	}

	if {[string length $errmsg] == 0 && $noprep == 0} {
		set hasfile 1
		set fd [open "pan.pre" r]
		while {[gets $fd line] > -1} {
		  if {[regexp -indices $pattern $line indices]} {
			catch "close $fd"
			return 1
		} }
		catch "close $fd"
		return 0
	}
	# else, ignore the include files:
	set noprep 1
	add_log "$errmsg"
	scan [.inp.t index end] %d numLines
	for {set i 1} {$i < $numLines} { incr i} {
		.inp.t mark set last $i.0
		if {[regexp -indices $pattern \
			[.inp.t get last "last lineend"] indices]} {
				return 1
		}
	}
	return 0
}
# xspin
# FSM view option

set nodeX(0) 0
set nodeY(0) 0
set Label(0) 0
set Transit(0) {}
set TLabel(0) 0
set Lab2Node(0) 0
set Dist(0) 0
set State(0) 0
set Edges(0) {}
set New 0
set MaxDist 0
set Maxx 0
set Maxy 0
set Minx 50
set Miny 50
set reached_end 0
set Scale 1
set Sml		"-*-Helvetica-Bold-R-Normal--*-60-*"

set cnr	0
set x 0
set y 0

proc fsmview {} {
	global Unix GO32

	add_log "<fsm view option>"

	if {[mk_exec] != 1} { return }

	catch {destroy .z}
	toplevel .z

	wm title .z "Double-Click Proc"

	listbox .z.list -setgrid 1
	button .z.b -text "Cancel" -command "destroy .z"

	pack append .z \
		.z.list {top expand fill} \
		.z.b {right padx 5}

	if {$Unix} {
		add_log "./pan -d	# find proc names"; update
		set fd [open "|./pan -d" w+]
	} else {
		add_log "$GO32 pan -d	# find proc names"; update
		catch { eval exec $GO32 pan -d >&pan.tmp } err
		if {$err != ""} {
			add_log "error: $err"
		}
		set fd [open pan.tmp r]
	}
	while {[gets $fd line] > -1} {
		if {[string first "proctype" $line] >= 0 } {
			.z.list insert end \
			[string trim [string range $line 9 end]]
	}	}
	catch { close $fd }

	bind .z.list <Double-Button-1> {
		set pfind [selection get]
		catch { destroy .z }
		findfsm $pfind
	}
	focus .z.list
}

proc findfsm {pfind} {
	global Unix GO32 New Dist State Edges Label
	global Transit MaxDist reached_end

	if {[mk_exec] != 1} { return }

	set src  0; set trn 0; set trg 0
	set Want 0; set MaxDist 0

	catch { foreach el [array names State] { unset State($el) } }
	catch { foreach el [array names Edges] { unset Edges($el) } }
	catch { foreach el [array names Dist]  { unset Dist($el) } }

	if {$Unix} {
		add_log "./pan -d	# compute fsm"; update
		set fd [open "|./pan -d" w+]
	} else {
		add_log "$GO32 pan -d	# compute fsm"; update
		catch { eval exec $GO32 pan -d >&pan.tmp }
		set fd [open pan.tmp r]
	}
	while {[gets $fd line] > -1} {
		set k [string first "proctype" $line]
		if { $k >= 0 } {
			if { $Want == 1 } {
				break
			}
			incr k 9
			set pname [string range $line $k end]
			if { [string first $pfind $line] >= 0 } {
				set Want 1
				set reached_end 0
				set nsrc "$pname:0"
				set Dist($nsrc) 0
				set Label($nsrc) "end"
				set Edges($nsrc) {}
			}
		} elseif { $Want == 1 \
			&& [string first "state" $line] >= 0} {
			scan $line "	state %d -(tr %d)-> state %d" \
				src trn trg
			if {$trg == 0} { set reached_end 1 }

			set nsrc "$pname:$src"
			set ntrg "$pname:$trg"
			set k [string first "line" $line]
			if { $k > 0 } {
				set m [string first "=>" $line]
				incr m -1
				set lbl [string range $line $k $m]
				incr m 3
				set action [string range $line $m end]
			} else {
				set lbl "line 0"
				set action "line 0"
			}
			set Label($nsrc) $lbl
			if { [info exists Dist($nsrc)] == 0 } {
				set Dist($nsrc) 0
				set Edges($nsrc) {}
			}
			if { [info exists Dist($ntrg)] == 0 } {
				set Dist($ntrg) [expr $Dist($nsrc) + 1]
				set Edges($ntrg) {}
				if {$Dist($ntrg) > $MaxDist } {
					set MaxDist $Dist($ntrg)
				}
			} else {
				if { [expr $Dist($nsrc) + 1] < $Dist($ntrg) } {
					set Dist($ntrg) [expr $Dist($nsrc) + 1]
					if {$Dist($ntrg) > $MaxDist } {
						set MaxDist $Dist($ntrg)
					}
			}	}
			lappend Edges($nsrc) $ntrg
			lappend Transit($nsrc) $action
		}
	}
	if { $Want == 1 } {
		dograph $pfind
	} else {
		add_log "sorry, $pfind not found..."
	}
	catch "close $fd"
	add_log "<done>"
	update
}

proc plunknode {el prefix} {
	global State Label
	global x y

	set pn [string range $el $prefix end]

	set State($el) [mkNode $Label($el) $pn $x $y]
	if { $x > 250 } {
		set x [expr $x - 250]
		set x [expr 250 - $x]
	} else {
		set x [expr 250 - $x]
		incr x 75
		set x [expr 250 + $x]
	}
}

proc dograph {n} {
	global Dist Edges Label Transit MaxDist State ELabel
	global lcnr reached_end x y Unix
	set count -1

	set lcnr [mkcanvas "FSM $n" $n 300 300 0]
	set prefix [string length $n]
	incr prefix
	set y 0

	while {$count < $MaxDist} {
		incr count
		incr y 50
		set x 250
		foreach el [array names Dist] {
			if { [ string first $n $el ] >= 0 \
			&&   $Dist($el) == $count \
			&&   $el != "$n:0" } {
				plunknode $el $prefix
		}	}
	}
	if {$reached_end == 1} {
		# force end state at the bottom
		incr y 50
		set x 250
		plunknode "$n:0" $prefix
	}

	foreach el [array names Edges] {
		if { [ string first $n $el ] >= 0 } {
			for {set i 0} { [lindex $Edges($el) $i] != "" } {incr i} {
				set ntrg [lindex $Edges($el) $i]
				set lbl  [lindex $Transit($el) $i]
				mkEdge $lbl $State($el) $State($ntrg)
				set ELabel($el,$ntrg) "$lbl"
			}
	}	}
	addscales $lcnr

	if {$Unix == 1 && [auto_execok dot] != 0} {
		button .f$lcnr.b6 -text "Redraw with Dot"  -command "dodot $n"
		pack append .f$lcnr .f$lcnr.b6 {right padx 5}
	}
	update
}

proc mkNode {n t x y} {
	# tcl book p. 213
	global cnr Sml NodeWidth NodeHeight
	global nodeX nodeY New TLabel Lab2Node

	set New [.f$cnr.c create oval [expr $x-10] [expr $y-10] \
		[expr $x+10] [expr $y+10] \
		-outline black \
		-fill white \
		-tags node]
	set Lab [.f$cnr.c create text $x $y -text $t -tags node]

	.f$cnr.c bind $Lab <Any-Enter> "
		.f$cnr.c itemconfigure $Lab -fill black -text {$n}
		if {[string first \"end\" $n] < 0 } { set_src {$n} }
	"
	.f$cnr.c bind $Lab <Any-Leave> "
		.f$cnr.c itemconfigure $Lab -fill black -text $t
	"

	set nodeX($New) $x
	set nodeY($New) $y
	set TLabel($New) $Lab
	set Lab2Node($Lab) $New
	set NodeWidth($New) 10
	set NodeHeight($New) 10

	update
	return $New
}

proc set_src {n} {
	set where 0
	scan $n "line %d" where
	.inp.t tag remove hilite 0.0 end
	src_line $where
}

proc mkEdge {L a b} {
	global cnr Xrem Yrem TransLabel Sml
	global nodeX nodeY edgeHead edgeTail

	if { $nodeY($b) > $nodeY($a) } {
		set ydiff -11
	} elseif { $nodeY($b) < $nodeY($a) } {
		set ydiff 11
	} else {
		set ydiff 0
	}
	if { $nodeX($b) > $nodeX($a) } {
		set xdiff -6
	} elseif { $nodeX($b) < $nodeX($a) } {
		set xdiff 6
	} else {
		set xdiff 0
	}
	set edge [.f$cnr.c create line \
		$nodeX($a) $nodeY($a) \
		[expr $nodeX($b) + $xdiff] \
		[expr $nodeY($b) + $ydiff] \
		 -arrow both -arrowshape {4 3 3} ]
	.f$cnr.c lower $edge
	lappend edgeHead($a) $edge
	lappend edgeTail($b) $edge

	set Xrem($edge) $xdiff
	set Yrem($edge) $ydiff

	set midX [expr $nodeX($a) + $nodeX($b)]
	set midX [expr $midX / 2 ]
	set midY [expr $nodeY($a) + $nodeY($b)]
	set midY [expr $midY / 2 ]

	set TransLabel($a,$b) [.f$cnr.c create text $midX $midY -font $Sml]

	.f$cnr.c bind $edge <Button-1> "
		.f$cnr.c coords $TransLabel($a,$b) \[.f$cnr.c canvasx %x\] \[.f$cnr.c canvasy %y\]
		.f$cnr.c itemconfigure $TransLabel($a,$b) \
			-fill black -text {$L} -font $Sml
	"
	.f$cnr.c bind $edge <ButtonRelease-1> "
		.f$cnr.c itemconfigure $TransLabel($a,$b) \
			-fill black -text {}
	"
}

proc moveNode {cnr node mx my together} {
	global edgeHead edgeTail TLabel NodeWidth NodeHeight
	global nodeX nodeY Lab2Node
	global Xrem Yrem Scale

	set x [.f$cnr.c coords $node]
	if {[llength $x] == 2} { set node $Lab2Node($node) }

	set mx [.f$cnr.c canvasx $mx]
	set my [.f$cnr.c canvasy $my]

	set wx $NodeWidth($node)
	set wy $NodeHeight($node)

	.f$cnr.c coords $node \
		[expr $mx-$wx] [expr $my-$wy] \
		[expr $mx+$wx] [expr $my+$wy]
	.f$cnr.c coords $TLabel($node) $mx $my

	set nodeX($node) $mx
	set nodeY($node) $my
	if {$together == 0} { return }
	catch {
	foreach edge $edgeHead($node) {
		set ec [.f$cnr.c coords $edge]
		set nx [expr $nodeX($node) + $Xrem($edge)*$Scale]
		set ny [expr $nodeY($node) + $Yrem($edge)*$Scale]
		.f$cnr.c coords $edge \
			$nx $ny \
			[lindex $ec 2] [lindex $ec 3]
	}}
	catch {
	foreach edge $edgeTail($node) {
		set ec [.f$cnr.c coords $edge]
		set nx [expr $nodeX($node) + $Xrem($edge)*$Scale]
		set ny [expr $nodeY($node) + $Yrem($edge)*$Scale]
		.f$cnr.c coords $edge \
			[lindex $ec 0] [lindex $ec 1] \
			$nx $ny
	}}
}

proc mkcanvas {n m geox geoy placed} {
	global cnr tk_version
	global Chandle Sticky
	global FG BG

	incr cnr
	toplevel .f$cnr
	wm title .f$cnr "$n"
	wm iconname .f$cnr $m
	if {$placed} {
		wm geometry .f$cnr +$geox+$geoy
	}
	wm minsize .f$cnr 100 100

	if {[string first "4." $tk_version] == 0 \
	||  [string first "8." $tk_version] == 0} {
		set cv [canvas .f$cnr.c  -relief raised -bd 2\
			-scrollregion {-15c -5c 30c 40c} \
			-background $BG \
			-xscrollcommand ".f$cnr.hscroll set" \
			-yscrollcommand ".f$cnr.vscroll set"]
		scrollbar .f$cnr.vscroll -relief sunken \
			-command ".f$cnr.c yview"
		scrollbar .f$cnr.hscroll -relief sunken -orient horiz \
			-command ".f$cnr.c xview"
	} else {
		set cv [canvas .f$cnr.c  -relief raised -bd 2\
			-scrollregion {-15c -5c 30c 40c} \
			-background $BG \
			-xscroll ".f$cnr.hscroll set" \
			-yscroll ".f$cnr.vscroll set"]
		scrollbar .f$cnr.vscroll -relief sunken \
			-command ".f$cnr.c yview"
		scrollbar .f$cnr.hscroll -relief sunken -orient horiz \
			-command ".f$cnr.c xview"
	}
	set Chandle($cnr) $cv
	#-width 500 -height 500

	button .f$cnr.b1 -text "Close" \
		-command "destroy .f$cnr"
	button .f$cnr.b2 -text "Save in: $m.bw.ps" \
		-command "$cv postscript -file $m.bw.ps -colormode mono"
	button .f$cnr.b3 -text "Save in: $m.col.ps" \
		-command "$cv postscript -file $m.col.ps -colormode color"
#	or use:	-colormode gray

	pack append .f$cnr \
		.f$cnr.vscroll {right filly} \
		.f$cnr.hscroll {bottom fillx} \
		.f$cnr.c {top expand fill}

	if {$m == "msc"} {
		set Sticky($cnr) 0
		checkbutton .f$cnr.b6 -text "Preserve" \
			-variable Sticky($cnr) \
			-relief flat
		pack append .f$cnr \
			.f$cnr.b6 { right padx 5}
	}
	pack append .f$cnr \
		.f$cnr.b1 {right padx 5} \
		.f$cnr.b2 {right padx 5} \
		.f$cnr.b3 {right padx 5}

	bind $cv <2> "$cv scan mark %x %y"	;# Geert Janssen, TUE
	bind $cv <B2-Motion> "$cv scan dragto %x %y"

	.f$cnr.c bind node <B1-Motion> "
		moveNode $cnr \[.f$cnr.c find withtag current] %x %y 1
	"

#	.f$cnr.c bind node <B2-Motion> "
#		moveNode $cnr \[.f$cnr.c find withtag current] %x %y 1
#	"

	tkwait visibility .f$cnr
	return $cnr
}

proc addscales {p} {
	global Chandle Scale

	catch {
		set cv $Chandle($p)
		button .f$p.b4 -text "Larger" \
			-command "$cv scale all 0 0 1.1 1.1; set Scale [expr $Scale*1.1]"
		button .f$p.b5 -text "Smaller" \
			-command "$cv scale all 0 0 0.9 0.9; set Scale [expr $Scale*0.9]"
		pack append .f$p \
			.f$p.b4 {right padx 5} \
			.f$p.b5 {right padx 5}
	}
}

proc dodot {n} {
	global Edges edgeHead edgeTail NodeWidth NodeHeight Maxx Maxy
	global State ELabel TransLabel Unix cnr lcnr Xrem Yrem

	if {$Unix == 0} {
		add_log "sorry, redraw requires dot"
		return
	}

	add_log "<dot graph layout...>"
	set fd [open "|dot" w+]

	puts $fd "digraph dodot {"

	foreach el [array names Edges] {
		if { [ string first $n $el ] >= 0 } {
		for {set i 0} { [lindex $Edges($el) $i] != "" } {incr i} {
			set ntrg [lindex $Edges($el) $i]
			puts $fd " \"$el\" -> \"$ntrg\"; "
		}
	}}

	puts $fd "}"
	flush $fd
	set ends ""
	set nodot 1
	while {[gets $fd line] > -1} {
		if {[string first "\}" $line] >= 0} {
			break;
		}
		set dd [string length $line]; incr dd -1
		while {[string range $line $dd $dd] == "\\"} {
			gets $fd more
			set line "[string range $line 0 [expr $dd-1]]$more"
			set dd [string length $line]; incr dd -1
		}
		if {[string first " -> " $line] >= 0} {
			set a [string first "\[pos=\"" $line]; incr a 8
			set b [string first "\"\];" $line]; incr b -1
			set b2 [string first "->" $line]
			set line1 [string range $line 0 [expr $a - 9]]
			set src [string range $line1 0 [expr $b2 - 1]]
			set src [string trim $src "	"]
			set src [string trim $src " "]
			set src [string trim $src "\""]
			set trg [string range $line1 [expr $b2 + 3] [expr $a - 1]]
			set trg [string trim $trg "	"]
			set trg [string trim $trg " "]
			set trg [string trim $trg "\""]
			set tp [string range $line [expr $a-2] [expr $a-2]]
			set line [string range $line $a $b]
			set k [split $line " "]
			set j [llength $k]
			set j2 [trunc [expr $j/2]]
			set coords ".f$cnr.c create line "
			set spline "-smooth 1"
			set nl $ELabel($src,$trg)
			if {$tp == "e"} {
				set ends "last"
				for {set i 1} {$i < $j} {incr i} {
					scan [lindex $k $i] "%d,%d" x y
					set coords " $coords[expr 50 + $x] [expr 50 + $Maxy - $y] "
					if {$i == $j2} {
						.f$cnr.c coords \
							$TransLabel($State($src),$State($trg)) \
							[expr 50 + $x] [expr 50 + $Maxy - $y]
						.f$cnr.c itemconfigure \
							$TransLabel($State($src),$State($trg)) \
							-fill black -text "$nl"
					}
				}
				scan [lindex $k 0] "%d,%d" x y
				set coords " $coords[expr 50 + $x] [expr 50 + $Maxy - $y] "
			} else {
				set ends "first"
				for {set i 0} {$i < $j} {incr i} {
					scan [lindex $k $i] "%d,%d" x y
					set coords " $coords[expr 50 + $x] [expr 50 + $Maxy - $y] "
					if {$i == $j2} {
						.f$cnr.c coords \
							$TransLabel($State($src),$State($trg)) \
							[expr 50 + $x] [expr 50 + $Maxy - $y]
						.f$cnr.c itemconfigure \
							$TransLabel($State($src),$State($trg)) \
							-fill black -text "$nl"
			}	}	}
			set coords "$coords -arrow $ends $spline -tags connector"

			set ne [eval $coords]
			set Xrem($ne) 10
			set Yrem($ne) 10

			continue
		}
		if {[string first "node " $line] >= 0 \
		||  [string first "\{" $line]    >= 0} {
			continue
		}
		if {[string first "graph " $line] >= 0} {
			set a [string first "\"" $line]; incr a
			set b [string last  "\"" $line]; incr b -1
			set line [string range $line $a $b]
			set k [split $line ","]
			if {[llength $k] == 4} {
				set Maxx [lindex $k 2]
				set Maxy [lindex $k 3]
			} else {
				set Maxx [lindex $k 0]
				set Maxy [lindex $k 1]
			}
			set nodot 0
		} else {	;# a node
			set a [string first "\[" $line]; incr a
			set b [string last  "\]" $line]; incr b -1
			set line1 [string range $line 0 [expr $a - 2]]
			set line  [string range $line $a $b]
			set nn [string trim $line1 "	"]
			set nn [string trim $nn " "]
			set nn [string trim $nn "\""]

			set a [string first "pos=" $line]; incr a 5
			set b [string first "\"" [string range $line $a end]]
			set b [expr $a + $b - 1]
			set line1  [string range $line $a $b]
			set k [split $line1 ","]
			set x [lindex $k 0]
			set y [lindex $k 1]

			set a [string first "width=" $line]; incr a 7
			set b [string first "\"" [string range $line $a end]]
			set b [expr $a + $b - 1]
			set w [expr 75 * [string range $line $a $b]]

			set a [string first "height=" $line]; incr a 8
			set b [string first "\"" [string range $line $a end]]
			set b [expr $a + $b - 1]
			set h [expr 75 * [string range $line $a $b]]

			set NodeWidth($State($nn)) [expr $w/2]
			set NodeHeight($State($nn)) [expr $h/2]
			moveNode $lcnr $State($nn) \
				[expr 50 + $x] [expr 50 + $Maxy - $y] 0
		}
	}
	catch { close $fd }

	if {$nodot} {
		add_log "<cannot find dot>"
		return
	}

	foreach el [array names Edges] {
		if { [ string first $n $el ] >= 0 } {
			catch {
			foreach edge $edgeHead($State($el)) {
				.f$cnr.c delete $edge
			}
			unset edgeHead($State($el))
			unset edgeTail($State($el))
			}
	}	}
	.f$cnr.c bind node <B1-Motion> {}	;# no moving
	.f$cnr.c bind node <B2-Motion> {}
	destroy .f$lcnr.b6
}

proc trunc {p} {
	set foo [string first "." $p]
	if {$foo >= 0} {
		incr foo -1
		set p [string range $p 0 $foo]
	}
	return $p
}
# xspin
# Help menu

proc helper {} {
	global FG BG

	catch {destroy .h}
	toplevel .h

	wm title .h "Help"
	wm iconname .h "Help"
	message .h.t -background $BG -foreground $FG \
	-text "\
Spin Version 3.0 Controller - (c) 1993-1997 Bell Laboratories

To use this interface you need Spin Version 2.8 or later.
You can obtain the latest version of Spin via anonymous ftp \
from netlib.bell-labs.com from directory /netlib/spin

Enter a Promela verification model into the text window,
or 'Open' one via the File Menu. \
Once loaded, you can revert to the stored version of the file \
with option ReOpen.  Select Clear to empty the text window.

On the standard output all Unix/shell commands \
are printed that the spin controller synthesizes. \
The output of Simulations or Verifications will \
appear in separate windows - with typically \
one window per running process.

The recommended way of working is to use Random Simulation for \
quick debugging (syntax errors are printed in the Log-Window and \
marked in the text of the main file window) \
- use Verification for real correctness proofs - and replay \
any errors found there with the Guided Simulation option.

You can select text fragments by sweeping the mouse with \
button-1 (the left-most button) down, and you can erase \
a selected fragment by hitting ctl-d (control-d).
To copy a selected text fragment, place the cursor at the
point of insertion and hit ctl-v (control-v)."
	button .h.b -text "Ok" -command {destroy .h}
	pack append .h .h.t {top expand fill}
	pack append .h .h.b {top}
}
# xspin
# LTL interface

set formula ""
set tl_stat 0

proc put_template {s} {
	.tl.main.e1 delete 0 end
	.tl.main.e1 insert end "$s"
}

proc call_tl {} {
	global formula tl_stat
	global FG BG

	catch {destroy .tl}
	toplevel .tl

	wm title .tl "Linear Time Temporal Logic Formulae"
	wm iconname .tl "LTL"
	wm geometry .tl +150+150
	frame .tl.main
	entry .tl.main.e1 -relief sunken -width 30 \
		-background $BG -foreground $FG
	label .tl.main.e2 -text "Formula: "

	frame .tl.op
	set alw {\[\] }
	set eve {\<\> }
	pack append .tl.op [label .tl.op.s0 -text "Operators: " \
		-relief flat] {left}
	pack append .tl.op [button .tl.op.always -text "\[\]" \
		-command ".tl.main.e1 insert end \"$alw \""] {left}
	pack append .tl.op [button .tl.op.event -text "\<\>" \
		-command ".tl.main.e1 insert end \"$eve \""] {left}
	pack append .tl.op [button .tl.op.until -text "U" \
		-command ".tl.main.e1 insert end \" U \""] {left}
	pack append .tl.op [button .tl.op.impl -text "->" \
		-command ".tl.main.e1 insert end \" -> \""] {left}
	pack append .tl.op [button .tl.op.and -text "and" \
		-command ".tl.main.e1 insert end \" && \""] {left}
	pack append .tl.op [button .tl.op.or -text "or" \
		-command ".tl.main.e1 insert end \" || \""] {left}
	pack append .tl.op [button .tl.op.not -text "not" \
		-command ".tl.main.e1 insert end \" ! \""] {left}

	frame .tl.tp -relief raised
	pack append .tl.tp [label .tl.tp.s0 -text "Templates: " \
		-relief flat] {left}
	pack append .tl.tp [button .tl.tp.inv \
		-text "Invariance(p)" \
		-command "put_template {\[] ( p )}"] {left}
	pack append .tl.tp [button .tl.tp.resp \
		-text "Response(p,q)" \
		-command "put_template {\[] ((p) -> (<> (q)))}"] {left}
	pack append .tl.tp [button .tl.tp.prec \
		-text "Precedence(p,q,r)" \
		-command "put_template {\[] ((p) -> ((q) U (r)))}"] {left}
	pack append .tl.tp [button .tl.tp.obj \
		-text "Objective(p,q,r)" \
		-command "put_template {\[] ((p) -> <> ((q) || (r)))}"] {left}

	frame .tl.b -relief ridge -borderwidth 4
	label .tl.b.s0 -text "This Property Is: "
	radiobutton .tl.b.s1 -text "Always Satisfied" \
		-variable tl_stat -value 0
	radiobutton .tl.b.s2 -text "Never Satisfied" \
		-variable tl_stat -value 1
	pack append .tl.b \
		.tl.b.s0 {top left} \
		.tl.b.s1 {top left} \
		.tl.b.s2 {top left}

	.tl.main.e1 insert end $formula

	menubutton .tl.main.file -text "Options.." \
		-relief raised -menu .tl.main.file.m

	menu .tl.main.file.m
	.tl.main.file.m add command -label "Help with LTL" \
		-command "roadmap4"
	.tl.main.file.m add command -label "Clear Formula" \
		-command ".tl.main.e1 delete 0 end"
	.tl.main.file.m add command -label "Clear Claim Window" \
		-command ".tl.never.t delete 0.0 end"
	.tl.main.file.m add command -label "Read Claim from File" \
		-command "browse_tl"
	.tl.main.file.m add command -label "Save Claim in File" \
		-command "save_tl"

	bind .tl.main.e1 <Return> { do_ltl }

	pack append .tl.main \
		.tl.main.e2 {top left}\
		.tl.main.e1 {top left expand fill} \
		.tl.main.file {top right}

	pack append .tl .tl.main {top fillx frame e}
	pack append .tl .tl.b {top fillx frame w}
	pack append .tl .tl.op {top frame w}
	pack append .tl .tl.tp {top frame w}

	frame .tl.never
	scrollbar .tl.never.s -relief flat \
		-command ".tl.never.t yview"
	text      .tl.never.t -relief raised -bd 2 \
		-yscrollcommand ".tl.never.s set" \
		-background $BG -foreground $FG \
		-setgrid 1 \
		-wrap word
	pack append .tl.never \
		.tl.never.s {left filly} \
		.tl.never.t {left expand fill}

	pack append .tl .tl.never {top expand fill}

	pack append .tl [button .tl.doit -text "Convert" \
		-command "do_ltl"] {right}
	pack append .tl [button .tl.imp -text "Accept" \
		-command "add_tl"] {right}
	pack append .tl [button .tl.sv -text "Save" \
		-command "save_tl"] {right}
	pack append .tl [button .tl.exit -text "Dismiss" \
		-command "destroy .tl"] {right}
#	pack append .tl [button .tl.help -text "Help" \
#		-command "roadmap4"] {right}
	focus .tl.main.e1
}

proc do_ltl {} {
	global formula tl_stat Unix SPIN tk_major tk_minor
	set formula "[.tl.main.e1 get]"
	.tl.never.t delete 0.0 end
	update

	if {$tl_stat == 0} {
		add_log "$SPIN -f \"!( $formula )\""
		if {$Unix || ($tk_major == 4 && $tk_minor >= 2)} {
		catch {exec $SPIN -f "!($formula)" >&pan.ltl} err
		} else {
		catch {exec $SPIN -f \"!($formula)\" >&pan.ltl} err
		}
	} else {
		add_log "$SPIN -f \" $formula\""
		if {$Unix} {
		catch {exec $SPIN -f " $formula" >&pan.ltl} err
		} else {
		catch {exec $SPIN -f \" $formula\" >&pan.ltl} err
		}
	}

	if {$err != ""} {
		add_log "<error: $err>"
		add_log "hint: check the Help Button for syntax rules"
	} else {
		.tl.never.t insert end \
		"	/*\n"
		.tl.never.t insert end \
		"	 * Formula As Typed: $formula\n"
		if {$tl_stat == 0} {
			.tl.never.t insert end \
			"	 * The Never Claim Below Corresponds\n"
			.tl.never.t insert end \
			"	 * To The Negated Formula !($formula)\n"
			.tl.never.t insert end \
			"	 * (formalizing violations of the original)\n"
		}
		.tl.never.t insert end \
		"	 */\n\n"
	}
	catch {
		set fd [open pan.ltl r]
		while {[gets $fd line] > -1} {
			.tl.never.t insert end "$line\n"
		}
		close $fd
	}
	rmfile pan.ltl
}

proc dump_tl {} {

	set fnm [.sv_tl.ent get]
	if {[file_ok $fnm]==0} return
	set fd [open $fnm w]
	add_log "<save claim in $fnm>"
	puts $fd "[.tl.never.t get 0.0 end]" nonewline
	catch "flush $fd"
	catch "close $fd"
	catch "destroy .sv_tl"
	catch "focus .tl.main.e1"
}

proc save_tl {} {
	catch {destroy .sv_tl}
	toplevel .sv_tl

	wm title .sv_tl "Save Claim"
	wm iconname .sv_tl "Save"
	wm geometry .sv_tl +200+200
	label  .sv_tl.msg -text "Name for the Claim File: " -relief flat
	entry  .sv_tl.ent -width 6 -relief sunken -textvariable fnm
	button .sv_tl.b1 -text "Ok" -command "dump_tl"
	button .sv_tl.b2 -text "Cancel" -command "destroy .sv_tl"
	bind   .sv_tl.ent <Return> { dump_tl }

	pack append .sv_tl \
		.sv_tl.msg {top frame w} \
		.sv_tl.ent {top frame e expand fill} \
		.sv_tl.b1 {right frame e} \
		.sv_tl.b2 {right frame e}
	focus .sv_tl.ent
}

proc add_tl {} {
	global BG FG
	catch {destroy .warn}
	toplevel .warn

	wm title .warn "Accept"
	wm iconname .warn "Accept"
	wm geometry .warn +200+200

	message .warn.t -width 300 \
		-background $BG -foreground $FG \
		-text " \
Instructions:

1. Save the Never Claim in a file, \
for instance a file called 'never', \
using the <Save> button.

2. Insert the line

#include \"never\"

(the name of the file with the claim) \
at the end of the main specification.

3. Insert macro definitions (#define's) for all \
propositional symbols used in the formula.

For instance, with LTL formula
'[] p -> <> q'  add the macro defs:

#define p	(cnt == 1)
#define q	(cnt > 1)

These macros must be defined just above the line \
with the #include \"never\" directive

4. Perform the verification, and make sure that \
the box 'Apply Never Claim' is checked in the \
Verification Panel (or else the claim is ignored).
You can have a library of claim files that you can \
choose from for verification, by changing only the \
name of the include file.

5. Never claims have no effect during simulation runs.

6. See the HELP->LTL menu for more information.

"
	button .warn.b -text "Ok" -command {destroy .warn}
	pack append .warn .warn.t {top expand fill}
	pack append .warn .warn.b {right frame e}
}

proc roadmap4 {} {
	global FG BG

	catch {destroy .h}
	toplevel .h

	wm title .h "LTL Help"
	wm iconname .h "Help"
	frame .h.z
	scrollbar .h.z.s -command ".h.z.t yview"
	text .h.z.t -relief raised -bd 2 \
		-background $BG -foreground $FG \
		-yscrollcommand ".h.z.s set" \
		-setgrid 1 -width 60 -height 30 -wrap word
	pack append .h.z \
		.h.z.s {left filly} \
		.h.z.t {left expand fill}
	.h.z.t insert end "
Define the LTL formula using lowercase proposition symbols,
for instance: [] (p U q)
The formula expresses either a positive (desired) or a
negative (undesired) property of the model.  A positive
property is negated automatically by the translator to
convert it in a never claim (which expresses the
corresponding negative property (the undesired behavior
that is claimed 'never' to occur).

When you type a <return> or hit the <Translate> button,
the formula is translated into an equivalent never-claim.

When you type the <Accept> button, you get instructions on
making the claim part of the model for verifications.

Finally, insert a macro-definition for each proposition symbol
used.  Insert these definitions above the never-claim in the
main text window. Enclose each definition in round braces,
to secure correct operator precedence, for instance:

#define p	(a > b)
#define q	(len(q) < 5)

Valid temporal logic operators are:
	\[\]  Always (no space between \[ and \])
	<>  Eventually (no space between < and >)
	U   (Strong) Until
	V   The Dual of Until: (p V q) == !(!p U !q)

	All operators are left-associative (incl. U and V).

Boolean Operators:
	&&  Logical And (alternative form: /\\, no spaces)
	!   Logical Negation
	||  Logical Or  (alternative form: \\/, no spaces)
	->  Logical Implication
	<-> Logical Equivalence

Boolean Predicates:
	true, false
	any name that starts with a lowercase letter

Examples:
	\[\] p
	!( <> !q )
	p U q
	p U (\[\] (q U r))

Generic properties:
	Invariance: \[\] p
	Response:   p -> \<\> q
	Precedence: p -> (q U r)
	Objective:  p -> \<\> (q || r)

	Each of the above 4 generic types of properties
	can (and will generally have to) be prefixed by
	temporal operators such as
		\[\], \<\>, \[\]\<\>, \<\>\[\]
	The last (objective) property can be read to mean
	that 'p' is a trigger, or 'enabling' condition that
	determines when the requirement becomes applicable
	(e.g. the sending of a new data message); then 'q'
	can be the fullfillment of the requirement (e.g.
	the arrival of the matching acknowledgement), and
	'r' could be a discharging condition that voids the
	applicability of the check (an abort condition).
"
	button .h.b -text "Ok" -command {destroy .h}
	pack append .h .h.z {top expand fill}
	pack append .h .h.b {top}
	.h.z.t configure -state disabled 
	.h.z.t yview -pickplace 1.0      
	focus .h.z.t
}


# xspin
# Specific Help

proc roadmap1 {} {
	global FG BG

	catch {destroy .road1}
	toplevel .road1

	wm title .road1 "Help with Simulation"
	wm iconname .road1 "SimHelp"
	frame .road1.z
	scrollbar .road1.z.s -command ".road1.z.t yview"
	text .road1.z.t -relief raised -bd 2 \
		-background $BG -foreground $FG \
		-yscrollcommand ".road1.z.s set" \
		-setgrid 1 -width 60 -height 30 -wrap word
	pack append .road1.z \
		.road1.z.s {left filly} \
		.road1.z.t {left expand fill}
	.road1.z.t insert end "
1.  Interactive simulation can be used to steer the\
execution towards a known point.  The user is prompted\
at every point in the execution where a nondeterministic\
choice has to be resolved (instead of using the random\
number generator to resolve the choices).

2.  Random simulation option can be used for debugging\
a validation model.  Other than assert statements, no\
correctness requirements are checked during simulation runs.\
You can skip an initial number of steps of the run using the\
Steps Skipped entry box in the Simulation Options Panel.

3.  During random simulation you can either do single-step\
execution, or an uninterrupted random run.  All nondeterministic\
decisions are now resolved randomly.  You can still force a selection\
to go a specific direction by modifying the model to force such\
choices.  A random run can be repeated precisely if the Seed Value\
for the random number generator is kept the same.\
You cannot go backwards in time. (See guided simulation below.)

4.  A guided simulation is used only to follow an error-trail that is\
produced by the verifier.  If the trail gets to be thousands of execution\
steps long, it can be time-consuming.  Make an effort, therefore to search\
for the shortest possible error-trail. (See verification help). \
You can again skip the initial portion of a long trail by typing a number in \
the box in the Simulation Panel that is labeled Steps Skipped.

5. In guided simulations you can go forwards or backwards in time.\
(But, beware that not all variable values are updated on backward steps.)\
The best mode is to let the simulation run to completion (the error state) and\
then look back in the traces a few steps to figure out the cause of the error.

6. The square option boxes in the Simulations Panel allow you to enable or\
disable types of listings and charts that will be maintained during simulation\
runs.  Usually, it is not necessary to look at all possible display panels.\
Experiment to see which displays you find most useful.

7. To track the value changes of selected variables, both in the \
Message Sequence Chart and in the Variable Values Panel, add the prefix \
'show ' to the declaration of the selected variables in the Promela \
specification, e.g. use 'show byte cnt;' instead of 'byte cnt;'"

	button .road1.b -text "Ok" -command {destroy .road1}
	pack append .road1 .road1.z {top expand fill}
	pack append .road1 .road1.b {top}
	.road1.z.t configure -state disabled 
	.road1.z.t yview -pickplace 1.0      
	focus .road1.z.t
}

proc roadmap2 {} {
	global FG BG

	catch {destroy .road2}
	toplevel .road2

	wm title .road2 "Help with Verification"
	wm iconname .road2 "ValHelp"
	frame .road2.z
	scrollbar .road2.z.s -command ".road2.z.t yview"
	text .road2.z.t -relief raised -bd 2 \
		-background $BG -foreground $FG \
		-yscrollcommand ".road2.z.s set" \
		-setgrid 1 -width 60 -height 30 -wrap word
	pack append .road2.z \
		.road2.z.s {left filly} \
		.road2.z.t {left expand fill}
	.road2.z.t insert end "
1. The Maximum Memory Usage option is used to define a hard\
upperbound to the amount of physical memory that you want the\
verifier to respect.  It the limit is reached, SPIN will\
automatically stop the verification.\
In case of doubt:  set it either to the exact amount of physical\
RAM in your system -- or slightly below that limit. \
It is interpreted to the nearest power of 2, so to increase simply\
double the number, to decrease it, half it.

2. The Estimated State Space Size is an independent parameter that\
is used to calculate the size of the hash-table.  There must be\
sufficient memory available (see above)\
to allocate a hash-table of the size desired.

3. Initially it is safe to leave all parameters at their\
default values and to try an exhaustive verification\
(simply hit the Run button).\
If you run out of memory, double the memory bound (Max Memory Usage),\
but make sure to *not exceed* the amount of physical (not virtual!)\
RAM that is available on your machine.

4. If you hit the maximum search depth (noted as `Search not completed' \
in the verification output window) without finding an error:\
double the search depth and repeat the run.

5. If an exhaustive verification cannot be completed due to lack\
of memory, switch to a supertrace analysis - and leave the memory bound\
at its maximum value.

6. Begin with the default estimate of the number of reachable states.\
After you successfully complete a run, double the estimate, and see\
if the number of reached stated changes much.  Continue to do this until\
it stops changing, or until you hit the maximum (i.e., larger than the\
memory bound you set earlier allows).

7a. If an error is found, try to reduce the search depth to try to find\
a shorter equivalent.  If you're content with the depth, move on to a\
guided simulation, to play back the error trail in detail.\
To find the shortest error-trail, continue to half the search depth\
limit until the error disappears.  Then continue to half the interval\
between the last level at which you saw the error, and the last level\
you checked where it no longer appeared.

7b. You can also use the Find Shortest Trail option to automate the\
process outlined in step 7a -- but it can be very expensive.  \
Do not use this option until you are certain that there exists an \
error -- leave the option turned off by default.

8. If searching for cycles -- use the fairness restriction only when\
really necessary.  It can increase the CPU-time requirements\
by a factor roughly equal to twice the number of active processes.

9. It is safe to leave the Reduced Search option set for most verifications.\
Turn it off only for debugging purposes, or when warned to do so by the\
verifier itself.  The Verbose option, similarly, is\
only for debugging, not for real verification work.  The Profiling option\
gathers statistics about the verification hot-spots in the model."

	button .road2.b -text "Ok" -command {destroy .road2}
	pack append .road2 .road2.z {top expand fill}
	pack append .road2 .road2.b {top}
	.road2.z.t configure -state disabled 
	.road2.z.t yview -pickplace 1.0      
	focus .road2.z.t
}

proc roadmap3 {} {
	global FG BG

	catch {destroy .road3}
	toplevel .road3

	wm title .road3 "Reducing Complexity"
	wm iconname .road3 "CompHelp"
	frame .road3.z
	scrollbar .road3.z.s -command ".road3.z.t yview"
	text .road3.z.t -relief raised -bd 2 \
		-background $BG -foreground $FG \
		-yscrollcommand ".road3.z.s set" \
		-setgrid 1 -width 60 -height 30 -wrap word
	pack append .road3.z \
		.road3.z.s {left filly} \
		.road3.z.t {left expand fill}
	.road3.z.t insert end "
When a verification cannot be completed because of\
computational complexity; here are some strategies that\
can be applied to combat this problem.

1. First, try to make the model more general, more abstract.\
Remember that you are constructing a verification model and not\
an implementation.  SPIN's strength is in proving properties of\
the Interactions in a distributed system (the implicit assumptions\
that processes make about each other) -- it's strength is not in\
proving things about local Computations, Data dependencies etc.

2. Remove everything that isn't directly related to the property\
you are trying to prove: redundant computations, redundant data. \
Avoid counters; avoid incrementing variables that are used for \
bookkeeping purposes.

3. Asynchronous channels are a notorious source of complexity in\
verifications.  Use a synchronous channel if possible.  Reduce the\
number of slots in asynchronous channels to a minimum (use 2, or 3\
slots to get started).

4. Combine local computations into atomic, or d_step, sequences.

5. Avoid leaving scratch data around in variables.  You can reduce\
the number of states by, for instance, resetting local variables\
that are used inside atomic sequences to zero at the end of those\
sequences; so that the scratch values aren't visible outside the\
sequence.  Alternatively: introduce some extra global `hidden'\
variables for these purposes (see the Doc/WhatsNew.ps document).
Use the predefined variable \"_\" as a write-only scratch variable\
wherever possible.

6. If possible to do so: combine the behavior of two processes into\
a single one.  Generalize behavior;  focus on coordination aspects\
(i.e., the interfaces between processes, rather than the local\
computation inside processes).

7. Try to exploit the partial order reduction strategies more.\
Use the xr and xs assertions (see Doc/WhatsNew.ps); avoid sharing\
channels between multiple receivers or multiple senders.\
Avoid sending independent data-streams through a shared channel."

	button .road3.b -text "Ok" -command {destroy .road3}
	pack append .road3 .road3.z {top expand fill}
	pack append .road3 .road3.b {top}
	.road3.z.t configure -state disabled 
	.road3.z.t yview -pickplace 1.0      
	focus .road3.z.t
}

proc roadmap5 {} {
	global FG BG

	catch {destroy .road5}
	toplevel .road5

	wm title .road5 "FSM Views"
	wm iconname .road5 "FsmHelp"
	frame .road5.z
	scrollbar .road5.z.s -command ".road5.z.t yview"
	text .road5.z.t -relief raised -bd 2 \
		-background $BG -foreground $FG \
		-yscrollcommand ".road5.z.s set" \
		-setgrid 1 -width 60 -height 30 -wrap word
	pack append .road5.z \
		.road5.z.s {left filly} \
		.road5.z.t {left expand fill}
	.road5.z.t insert end "
The Fsm view panels give a quick view of the precise
structure of the Finite State Machine models that Spin
uses in the verification process.  Each proctype is
represented by one unique FSM.

Clicking the Fsm View button in the Xspin display will
cause Spin to first generate a verifier, compile it, and
then run it (as pan -d) to obtain a dump of the proctype
names and state machines.

After selecting (double-clicking) the proctype name desired,
the graph will be produced.  The graph layout algorithm is
small and a self-contained part of Xspin, but also rather
crude.  Be on guard, therefore, for edges that
overlap (a typical case, for instance, is a backedge that
hides behind a series of forward edges.
Move nodes around with Button-1 or Button-2 to discover
such unpleasantries.  If needed - you can also generate the
FSM descriptions by hand (cc pan.c; pan -d) and use a small
awk-script or perl-script to translate to the input of more
elegant graph layout tools (e.g., dot).

1. Moving Nodes: either Button-1 or Button-2.

2. Displaying Edge Labels: hold Button-1 down on the edge.

3. Cross-References: Move the cursor over a Node to see the
corresponding line in the Promela source, in the main
Xspin window.
"
	button .road5.b -text "Ok" -command {destroy .road5}
	pack append .road5 .road5.z {top expand fill}
	pack append .road5 .road5.b {top}
	.road5.z.t configure -state disabled 
	.road5.z.t yview -pickplace 1.0      
	focus .road5.z.t
}
# xspin
# simulation options panel

set s_options	""
set v_options	""
set a_options	""
set c_options	""

set Blue	"blue";		#"yellow"
set Yellow	"yellow";	#"red"
set White	"white";	#"yellow"
set Red		"red";		#"yellow"
set Green	"green";	#"green"

set SmallFont	"-*-Courier-Bold-R-Normal--*-60-*"
set BigFont	"-*-Courier-Bold-R-Normal--*-120-*"

set fd		0
set Depth	0
set Seq(0)	0
set Sdbox	0
set Spbox(0)	0
set sbox	0

set simruns	0
set stepper	0
set stepped	0
set VERBOSE	0
set SYMBOLIC	0
set howmany	0
set Choice(1)	0
set Sticky(0)	0
set SparseMsc	1
set allvars	-1

proc simulation {} {
	global s_typ l_typ stop allvars hasfile
	global fvars gvars lvars SparseMsc
	global msc ebc tsc vv svars rvars seed jumpsteps

	set stop 0
	set hasfile 0

	catch {destroy .s}
	toplevel .s

	wm title .s "Simulation Options"
	wm iconname .s "SIM"
	wm geometry .s +150+150

	frame .s.opt -relief flat

	mkpan_in

	if {$allvars == -1} {
		if {[hasWord "show"]} {
			set allvars 0
			set vv 1
		} else {
			set allvars 1
		}
	}

	frame .s.opt.mode -relief raised -borderwidth 1m
	label .s.opt.mode.fld0 \
		-font -Adobe-Helvetica-Bold-R-Normal-*-120-* \
		-text "Display Mode" \
		-relief sunken -borderwidth 1m

	checkbutton .s.opt.mode.fld4b -text "Time Sequence Panel - with:" \
		-variable tsc \
		-relief flat
	radiobutton .s.opt.mode.fld3 -text "    Interleaved Steps" \
		-variable m_typ -value 2 \
		-relief flat
	radiobutton .s.opt.mode.fld1 -text "    One Window per Process" \
		-variable m_typ -value 0 \
		-relief flat
	radiobutton .s.opt.mode.fld2 -text "    One Trace per Process" \
		-variable m_typ -value 1 \
		-relief flat

	checkbutton .s.opt.mode.fld4a -text "MSC Panel - with:" \
		-variable msc \
		-relief flat
	radiobutton .s.opt.mode.fld5 -text "    Step Number Labels" \
		-variable SYMBOLIC -value 0 \
		-relief flat
	radiobutton .s.opt.mode.fld6 -text "    Source Text Labels" \
		-variable SYMBOLIC -value 1 \
		-relief flat

	radiobutton .s.opt.mode.fld7 -text "    Normal Spacing" \
		-variable SparseMsc -value 1 \
		-relief flat
	radiobutton .s.opt.mode.fld8 -text "    Condensed Spacing" \
		-variable SparseMsc -value 0 \
		-relief flat

	checkbutton .s.opt.mode.fld4c -text "Execution Bar Panel" \
		-variable ebc \
		-relief flat
	checkbutton .s.opt.mode.fld4d -text "Variable Values Panel" \
		-variable vv \
		-relief flat
	radiobutton .s.opt.mode.fld4e -text "    Track all variables" \
		-variable allvars -value 1 \
		-relief flat
	radiobutton .s.opt.mode.fld4f -text "    Selected variables (see Help 7.)" \
		-variable allvars -value 0 \
		-relief flat

	pack append .s.opt.mode .s.opt.mode.fld0 {top pady 4 frame w fillx}

	pack append .s.opt.mode .s.opt.mode.fld4a {top pady 4 frame w}
	pack append .s.opt.mode .s.opt.mode.fld5 {top pady 4 frame w}
	pack append .s.opt.mode .s.opt.mode.fld6 {top pady 4 frame w}
	pack append .s.opt.mode .s.opt.mode.fld7 {top pady 4 frame w}
	pack append .s.opt.mode .s.opt.mode.fld8 {top pady 4 frame w}

	pack append .s.opt.mode .s.opt.mode.fld4b {top pady 4 frame w}
	pack append .s.opt.mode .s.opt.mode.fld3 {top pady 4 frame w}
	pack append .s.opt.mode .s.opt.mode.fld1 {top pady 4 frame w}
	pack append .s.opt.mode .s.opt.mode.fld2 {top pady 4 frame w}

	pack append .s.opt.mode .s.opt.mode.fld4d {top pady 4 frame w}
	pack append .s.opt.mode .s.opt.mode.fld4e {top pady 4 frame w}
	pack append .s.opt.mode .s.opt.mode.fld4f {top pady 4 frame w}

	pack append .s.opt.mode .s.opt.mode.fld4c {top pady 4 frame w}

	pack append .s.opt .s.opt.mode {left frame n}

	frame .s.opt.mesg -relief raised -borderwidth 1m
	label .s.opt.mesg.loss0 \
		-font -Adobe-Helvetica-Bold-R-Normal-*-120-* \
		-text "A Full Queue" \
		-relief sunken -borderwidth 1m
	radiobutton .s.opt.mesg.loss1 -text "Blocks New Msgs" \
		-variable l_typ -value 0 \
		-relief flat
	radiobutton .s.opt.mesg.loss2 -text "Loses New Msgs" \
		-variable l_typ -value 1 \
		-relief flat
	pack append .s.opt.mesg .s.opt.mesg.loss0 {top pady 4 frame w fillx}
	pack append .s.opt.mesg .s.opt.mesg.loss1 {top pady 4 frame w}
	pack append .s.opt.mesg .s.opt.mesg.loss2 {top pady 4 frame w}

	frame .s.opt.lframe -relief raised -borderwidth 1m
	label .s.opt.lframe.tl \
		-font -Adobe-Helvetica-Bold-R-Normal-*-120-* \
		-text "Simulation Style" \
		-relief sunken -borderwidth 1m
	radiobutton .s.opt.lframe.is -text "Interactive" \
		-variable s_typ -value 2 \
		-relief flat
	radiobutton .s.opt.lframe.gs -text "Guided (using pan.trail)" \
		-variable s_typ -value 1 \
		-relief flat
	frame .s.opt.lframe.b
	entry .s.opt.lframe.b.entry -relief sunken -width 8 
	label .s.opt.lframe.b.label \
		-font -Adobe-Helvetica-Bold-R-Normal-*-120-* \
		-text "Steps Skipped"
	pack append .s.opt.lframe.b \
		.s.opt.lframe.b.label {left} \
		.s.opt.lframe.b.entry {left}
	radiobutton .s.opt.lframe.rs -text "Random (using seed)" \
		-variable s_typ -value 0 \
		-relief flat
	entry .s.opt.lframe.entry -relief sunken -width 8 
	label .s.opt.lframe.label \
		-font -Adobe-Helvetica-Bold-R-Normal-*-120-* \
		-text "Seed Value"

	pack append .s.opt.lframe .s.opt.lframe.tl {top pady 4 frame w fillx} \
		.s.opt.lframe.rs {top pady 4 frame w} \
		.s.opt.lframe.gs {top pady 4 frame w} \
		.s.opt.lframe.b  {top pady 4 frame e} \
		.s.opt.lframe.is {top pady 4 frame w} \
		.s.opt.lframe.entry right \
		.s.opt.lframe.label right

	pack append .s.opt .s.opt.lframe {top frame n}
	pack append .s.opt .s.opt.mesg {top frame w expand fillx filly}

	pack append .s .s.opt { top frame n }

	pack append .s [button .s.rewind -text "Start" \
		-command "Rewind"  ] {right frame e}
	pack append .s [button .s.exit -text "Cancel" \
		-command "Stopsim" ] {right frame e}
	pack append .s [button .s.help -text "Help" -fg red \
		-command "roadmap1" ] {right frame e}

	.s.opt.lframe.entry  insert end $seed
	.s.opt.lframe.b.entry insert end $jumpsteps
}

proc bld_s_options {} {
	global fvars gvars lvars svars
	global rvars l_typ allvars vv
	global s_typ seed jumpsteps s_options

	set s_options "-X -p -v"
	if {$allvars==0} { set s_options [format "%s -Y" $s_options] }
	if {$s_typ==2} { set s_options [format "%s -i" $s_options] }
	if {$vv && $gvars} { set s_options [format "%s -g" $s_options] }
	if {$vv && $lvars} { set s_options [format "%s -l" $s_options] }
	if {$svars} { set s_options [format "%s -s" $s_options] }
	if {$rvars} { set s_options [format "%s -r" $s_options] }
	if {$l_typ} { set s_options [format "%s -m" $s_options] }
	if {$s_typ==1} then {
		set s_options [format "%s -t" $s_options]
	} else {
		if {[string length $seed] > 0} {
			set s_options [format "%s -n%s" $s_options $seed]
	}	}
	if {$s_typ!=2} then {
		if {[string length $jumpsteps] > 0} {
			set s_options [format "%s -j%s" $s_options $jumpsteps]
	}	}
}

proc Stopsim {} {
	global stop sbox dbox2 Sticky
	global stepper stepped howmany

	set stop 1
	set stepped 0
	set stepper 0
	add_log "<stop simulation>"
	catch {destroy .s}
	catch {set howmany 0}
	catch {stopbar}
	catch {	if {$Sticky($dbox2) == 0} {
			destroy .f$dbox2
	}	}
}

proc Step_forw {} {
	global stepper stepped sbox simruns

	set stepped 1
	set stepper 1
	if {$simruns == 0} {
		catch { destroy .s }
		runsim
	} else {
		catch { .c$sbox.run configure \
		-text "Run" -command "Runsim" }
	}
}

proc Step_back {} {
	global Depth Seq fd

	set Depth [expr $Depth - 2]
	if {$Depth < 0} { set Depth 0 }
	catch { seek $fd $Seq($Depth) }
	Step_forw
}

proc Rewind {} {
	global simruns Depth
	global fd s_typ
	global Sdbox Spbox
	global seed jumpsteps

	catch { set jumpsteps [.s.opt.lframe.b.entry get] }

	if {$s_typ == 0} {
		set seed [.s.opt.lframe.entry get]
	}
	if {$s_typ == 1} {
		set Depth 0
		if {$simruns} { seek $fd 0 }
		catch {
			foreach el [array names Spbox] {
				set Sdbox $Spbox($el)
				.c$Sdbox.z.t tag remove Rev 1.0 end
	}	}	}
	Step_forw
}

proc Runsim {} {
	global stepper stepped sbox

	catch { .c$sbox.run configure \
		-text "Suspend" -command "Step_forw" }
	set stepper 1
	set stepped 0
}

proc BreakPoint {} {
	global stepped sbox

	set stepped 1
	catch { .c$sbox.run configure \
		-text "BreakPoint" -command "Runsim" }
}

proc runsim {} {
	global Unix SPIN GO32 tk_major
	global s_options s_typ dbox2
	global stepper stepped
	global simruns m_typ
	global gvars lvars
	global fd stop Depth Seq
	global Sdbox Spbox pbox howmany Choice
	global sbox VERBOSE SYMBOLIC msc ebc vv tsc
	global Blue Yellow White Red Green
	global SmallFont BigFont Sticky SparseMsc
	global FG BG

	set simruns 1
	set Vvbox 0
	set pno 0
	set Varnm("") ""
	set Queues("")	""
	set Depth 0
	set Seq(0) 0
	set Pstp 1
	set Seenpno 1
	set Banner "Select"

#	catch { unset Spbox(0) }
	catch {
		foreach el [array names pbox] {
			catch { destroy .c$pbox($el) }
			catch { unset pbox($el) }
		}
		foreach el [array names Spbox] {
			catch { destroy .c$Spbox($el) }
			catch { unset Spbox($el) }
		}
	}
	bld_s_options

	add_log "<starting simulation>"
	add_log "$SPIN $s_options pan_in"
	update
	set s_options [format "%s pan_in" $s_options]

	mkpan_in

	set sbox [mkbox "Simulation Output" "SimOut" "sim.out" 80 10 150 150]

	if {$s_typ == 1} {
		pack append .c$sbox [button .c$sbox.rewind -text "Step 1" \
		-command "Rewind" ] {left frame w}
		pack append .c$sbox [button .c$sbox.stepf -text "Step Forw" \
		-command "Step_forw" ] {left frame w}
		pack append .c$sbox [button .c$sbox.stepb -text "Step Back" \
		-command "Step_back" ] {left frame w}
	} else {
		pack append .c$sbox [button .c$sbox.stepf -text "Single Step" \
		-command "Step_forw" ] {left frame w}
	}
	pack append .c$sbox [button .c$sbox.run -text "Run" \
		-command "Runsim" ] {left frame w}

	.c$sbox.b configure -text "Cancel" -command "Stopsim"

	set YSZ 12
	set XSZ 84
	set YNR 60
	set NPR 10
	set SMX 250
	set Easy 1
	set HAS 0
	set HAS_CYCLE 0
	set dontwait 0
	set notexecutable 0
	set lastexecutable 0

	if {$m_typ == 2} {
		if {$tsc} {
		set pbox(0) \
		[mkbox "Time Sequence" "Sequence" "seq.out" 80 10 150 375]
		set dbox $pbox(0)
	}	}
	if {$msc} {
		if {[hasWord "!!"] || [hasWord "\\?\\?"]} {
			set Easy 0
		}

		set iniw 200
		set inih [expr $YNR*$YSZ+16]
		set maxx [expr [winfo screenwidth .]  - $iniw - 250]
		set maxy [expr [winfo screenheight .] - 50]
		if {$inih > $maxy} { set inih $maxy }
		set maxy [expr $maxy - $inih]

		set dbox2 \
		[mkcanvas "Message Sequence Chart" "msc" $maxx $maxy 1]
		.f$dbox2.c configure \
			-height $inih -width $iniw\
			-scrollregion "[expr -$XSZ/2] 0 \
				[expr $NPR*$XSZ] [expr $SMX*$YSZ]"
	}

	if {$s_typ == 1} {
		catch { .c$sbox.z.t insert end "preparing trail, please wait..." }
		update
		rmfile trail.out
		catch { eval exec $SPIN $s_options >&trail.out } errmsg
		if {[string length $errmsg]>0} {
			add_log "$errmsg"
			catch {
				set fd [open trail.out r]
				while {[gets $fd line] > -1} {
					add_log "$line"
				}
				close $fd
			}
			Stopsim
			catch { destroy .c$sbox }
			catch { destroy .c$dbox }
			set simruns 0
			update
			return
		}
		set fd [open trail.out r]
		catch { .c$sbox.z.t insert end "done\n" }
	} else {
		update
		set fd [open "|$SPIN $s_options" w+]
		catch "flush $fd"
		update
	}

	if {$s_typ == 2} {
		Runsim
	}

	if {$ebc} { startbar "Xspin Bar Chart" }

	set pstp -1
	set bailout 0

	while {$stop == 0 && [eof $fd] == 0} {
	if {$bailout == 0 && [gets $fd line] > -1} {
		set pln 0
		set syntax 0
		set isvar 0
		set pname ""
		set i 0
		set VERBOSE 0
		set Fnm "pan_in"
		if {$Unix == 0} {
			if {[string first "processes created" $line] > 0} {
				set bailout 1
		}	}

		set lastpstp $pstp
		set pmtch [scan $line \
			"%d: proc %d (%s line %d \"%s\" " \
			pstp pno pname pln Fnm]
		incr pmtch -1
		set i [string first "\[" $line]
		if {$i > 0} {
			set i [expr $i + 1]
			set j [string length $line]
			set j [expr $j - 2]
			set stmnt [string range $line $i $j]
		} else {
			set stmnt "-"
		}
		if {$pmtch != 4} {
			set pmtch [scan $line \
				"	proc %d (%s line %d \"%s\" " \
				pno pname pln Fnm]
		}
		if {$pmtch != 4} {
			if {[string first "spin: line" $line] == 0 } {
				scan $line "spin: line %d \"%s\" " pln Fnm
				if {[string first "pan_in" $Fnm] >= 0} {
				.inp.t tag add Rev $pln.0 $pln.end
				.inp.t tag configure Rev \
					-background $FG -foreground $BG
				.inp.t yview -pickplace $pln.0
				}
				if {[string first "assertion viol" $line] < 0} {
					set syntax 1
				}
			}
			if {[string first "Error: "   $line] >= 0 \
			||  [string first "warning: " $line] >= 0 } {
				set syntax 1
			}
		}
		if {$pmtch != 4 && $syntax == 0} {
			set pmtch [scan $line \
				"%d: proc - (%s line %d \"%s\" " \
				pstp pname pln Fnm]
			if { $pmtch == 4 } {
				set pno -1
			}
		}
		#	set Fnm [string trim $Fnm "\""]
		set pname [string trim $pname "()"]

		if {[string first "TRACK" $pname] >= 0} {
			set nwcol([expr $pno+1]) 1
		} elseif {[string length $pname] > 0} {
			if {[info exists nwcol([expr $pno+1])] \
			&&  $nwcol([expr $pno+1])} {
				unset Plabel($pno)
##
				set TMP1 [expr ($pno + 1)*$XSZ]
				set TMP2 [expr $Pstp*$YSZ]
				.f$dbox2.c create line \
					[expr $TMP1 - 20] $TMP2 \
					[expr $TMP1 + 20] $TMP2 \
					-width 2 \
					-fill $Red
				incr TMP2 4
				.f$dbox2.c create line \
					[expr $TMP1 - 20] $TMP2 \
					[expr $TMP1 + 20] $TMP2 \
					-width 2 \
					-fill $Red
##
			}
			set nwcol([expr $pno+1]) 0
		}
		if {$pmtch == 4 && $syntax == 0} {
			if {$ebc} {
				if {[string first "values:" $line] < 0} {
					stepbar $pno $pname
			}	}
			if {$m_typ == 1 && $tsc} {
				if { [info exists pbox($pno)] == 0 } {
					set pbox($pno) [mkbox \
						"Proc $pno ($pname)" \
						"Proc$pno" "proc.$pno.out" \
						60 10 \
						[expr 150+$pno*25] \
						[expr 375+$pno*35] ]
				}
				set dbox $pbox($pno)
			} elseif {$m_typ == 0 && $tsc} {
				if { [info exists Spbox($pno)] == 0 } {
					set Spbox($pno) \
						[mkbox "$pname (proc $pno)" \
						"$pname" "" \
						60 10 \
						[expr 150+$pno*25] \
						[expr 375+$pno*35] ]
					readinfile .c$Spbox($pno).z.t "pan_in"
				}
				set Sdbox $Spbox($pno)
			}
		} elseif { [string first "..." $line] > 0 && \
			   [string first "MSC: " $line] != 0} {
			set $line ""
			set syntax 1
			set pln 0
		} elseif {$s_typ == 2 \
		      && [string first "Select " $line] == 0 } {
			set Banner $line
			set pln 0
			set notexecutable 0
			set lastexecutable 0
			set has_timeout 0
		} elseif {$s_typ == 2 \
		      && [string first "choice" $line] >= 0 } {
			scan $line "	choice %d" howmany
			set NN [string first ":" $line]; incr NN 2
			set Choice($howmany) [string range $line $NN end]
			if {[string first "timeout" $Choice($howmany)] > 0} {
				set has_timeout 1
			}
			if {[string first "unexecutable," $line] >= 0} {
				incr notexecutable
			} else {
				set lastexecutable $howmany
			}
			set pln 0
		} elseif {$s_typ == 2 \
		      && [string first "Make Selection" $line] >= 0 } {
			scan $line "Make Selection %d" howmany
			if {$notexecutable == $howmany-1 && $has_timeout == 0} {
				set howmany $lastexecutable
				add_log "selected: $howmany (forced)"
			} else {
				pickoption $Banner
				add_log "selected: $howmany"
			}
			puts $fd $howmany
			catch "flush $fd"
			set dontwait 1
			set pln 0
		} elseif { [string first "MSC: " $line] == 0 } {
			set realstring [string range $line 5 end]
			# picked up in next cycle
		} elseif { [string first "processes" $line] > 0 \
		      ||   [string first "timeout" $line]  == 0 \
		      ||   [string first "=s==" $line]  > 0 \
		      ||   [string first "=r==" $line]  > 0 } {

			if { $m_typ == 1 && $tsc} {
				set dbox $pbox(0)
			} elseif { $m_typ == 0 && $tsc} {
				if { [info exists Spbox($pno)] == 0 } {
					set Spbox($pno) \
						[mkbox "$pname (proc $pno)" \
						"$pname" "" \
						60 10 \
						[expr 150+$pno*25] \
						[expr 375+$pno*35] ]
					readinfile .c$Spbox($pno).z.t "pan_in"
				}
				set Sdbox $Spbox($pno)
			}
			set pln 0;	# prevent tag update
		} elseif {$syntax == 0 && [string first " = " $line] > 0 } {
				set isvar [string first "=" $line]
				set isvar [expr $isvar + 1]
				set varvl [string range $line $isvar end]
				set isvar [expr $isvar - 2]
				set varnm [string range $line 0 $isvar]
				set varnm [string trim $varnm "	"]
				set Varnm($varnm) $varvl
				set isvar 1
		} elseif { [scan $line " %s %d " varnm qnr] == 2} {
			if {$syntax == 0 &&  [string compare $varnm "queue"] == 0} {
				set isvar [string last ":" $line]
				set isvar [expr $isvar + 1]
				set varvl [string range $line $isvar end]
				set XX [string first "(" $line]
				set YY [string last  ")" $line]
				set ZZ [string range $line $XX $YY]
				set Queues($qnr) $varvl
				if {[info exists Alias($qnr)]} {
				if {[string first $ZZ $Alias($qnr)] < 0} {
					set Alias($qnr) "$Alias($qnr), $ZZ"
				}
				} else {
					set Alias($qnr) $ZZ
				}
				set isvar 1
			}
		} elseif {[string length $line] == 0} {
			if {$dontwait == 0} { set stepper 0 }
			set pln 0
			set Depth [expr $Depth + 1]
			set Seq($Depth) [tell $fd]
			set dontwait 0
		}


	if {$syntax == 0} {
		if {[string first "terminates" $line] > 0} {
			set pln -1
			set stmnt "<stop>"
		}

		if {$m_typ == 0} {
			if {$pln != 0 && [string first "pan_in" $Fnm] >= 0} {
				catch {
				.c$Sdbox.z.t yview -pickplace $pln.0
				.c$Sdbox.z.t tag remove Rev 1.0 end
				.c$Sdbox.z.t tag add Rev $pln.0 $pln.end
				.c$Sdbox.z.t tag configure Rev \
					-background $FG -foreground $BG
			}	}
		} elseif {$m_typ == 1} {
			if { [info exists pbox($pno)] == 0 } {
				set pbox($pno) [mkbox \
					"Proc $pno ($pname)" \
					"Proc$pno" "proc.$pno.out" \
					60 10 \
					[expr 150+$pno*25] \
					[expr 375+$pno*35] ]
			}
			set dbox $pbox($pno)
			catch {
				.c$dbox.z.t yview -pickplace end
				.c$dbox.z.t insert end "$line\n"
			}
		} elseif {$m_typ == 2 && $pln != 0 \
		&&  [string first "unexecutable, " $line] < 0} {
			catch { .c$dbox.z.t yview -pickplace end }
			catch { .c$dbox.z.t insert end "$pno:$pln" }
			for {set i $pno} {$i > 0} {incr i -1} {
				catch { .c$dbox.z.t insert end "\t|" }
			}
			catch { .c$dbox.z.t insert end "\t|>$stmnt\n" }
		}
		if {$msc && $pln != 0} {
			set Mcont "--"
			set HAS 0
			if { [scan $stmnt "values: %d!%d" inq inp1] == 2 \
			||   [scan $stmnt "values: %d!%s" inq inp2] == 2 } {
				set HAS   [string first "!" $stmnt]
				incr HAS
				set Mcont [string range $stmnt $HAS end]
				set HAS 1
			} elseif { [scan $stmnt "values: %d?%d" inq inp1] == 2 \
			|| [scan $stmnt "values: %d?%s" inq inp2] == 2 } {
				set HAS   [string first "?" $stmnt]
				incr HAS
				set Mcont [string range $stmnt $HAS end]
				set HAS 2
			} elseif { [string first "-" $stmnt] == 0} {
				set HAS 3
				if {$HAS_CYCLE} {
					set stmnt [format "Cycle>>"]
				} else {
					set stmnt [format "<waiting>"]
				}
			} elseif { [string first "<stop>" $stmnt] == 0} {
				set HAS 3
				set stmnt [format "<stopped>"]
			}
			if {$pno+1 > $Seenpno} { set Seenpno [expr $pno+1] }
			set XLOC [expr (1+$pno)*$XSZ]
			set YLOC [expr $Pstp*$YSZ]
			if {[string first "printf('MSC: " $stmnt] == 0} {
				set VERBOSE 1
				set stmnt $realstring
				if {[string first "BREAK" $realstring] >= 0} {
					BreakPoint
				}
			} else {
				set VERBOSE 0
			}
			catch {
			if {$VERBOSE \
			||  $HAS != 0 \
			||  [info exists R($pstp,$pno)]} {

				if { $SparseMsc == 1 \
				||   [info exists Plabel($pno)] == 0 \
				||   ([info exists R($pstp,$pno)] == 0 \
				&&   ($HAS != 1 \
				||   [info exists HasBox($YLOC,[expr 1+$pno])])) } {
					incr Pstp
					for	{set i 1} \
						{$Pstp > 1 && $i <= $Seenpno} \
						{incr i} {
						if {[info exists HasBox($YLOC,$i)]} {
							continue
						}
						set TMP1 [expr $i*$XSZ]
set lncol $Blue
set lnwdt 1
catch {
	if {$nwcol($i)} {
		set lncol "gray"
		set lnwdt 15
}	}
						.f$dbox2.c create line \
						$TMP1 $YLOC $TMP1 \
						[expr $YLOC+$YSZ] \
						-width $lnwdt \
						-fill $lncol
					}
					if {[info exists HasBox($YLOC,[expr 1+$pno])]} {
						set YLOC [expr $Pstp*$YSZ]
				}	}
				if {$HAS == 1 || $HAS == 2} {
					set stmnt [string range $stmnt 8 end]
				}
				if { [info exists Plabel($pno)] == 0} {
					set Plabel($pno) 0
					if {$SparseMsc == 0} {
						set HasBox($YLOC,[expr 1+$pno]) 1
					}
					.f$dbox2.c create rectangle \
						[expr $XLOC-20] $YLOC \
						[expr $XLOC+20] \
						[expr $YLOC+$YSZ] \
						-outline $Red -fill $Yellow
				if {$pname != "TRACK"} {
					.f$dbox2.c create text $XLOC \
						[expr $YLOC+$YSZ/2] \
						-font $SmallFont \
						-text "$pname:$pno"
				} else {
					.f$dbox2.c create text $XLOC \
						[expr $YLOC+$YSZ/2] \
						-font $SmallFont \
						-text "<show>"
				}

					set YLOC [expr $Pstp*$YSZ]
					incr Pstp
					for	{set i 1} \
						{$Pstp > 1 && $i <= $Seenpno} \
						{incr i} {
						set TMP1 [expr $i*$XSZ]
				set lncol $Blue
				set lnwdt 1
				catch {
					if {$nwcol($i)} {
						set lncol "gray"
						set lnwdt 15
				}	}
						.f$dbox2.c create line \
						$TMP1 $YLOC $TMP1 \
						[expr $YLOC+$YSZ] \
						-width $lnwdt \
						-fill $lncol
					}
				}
				if {(1+$pno) > $NPR} {
					set NPR [expr $pno+2]
					.f$dbox2.c configure \
					 -scrollregion \
					 "[expr -$XSZ/2] 0 \
					  [expr $NPR*$XSZ] [expr $SMX*$YSZ]"
				}
				if {$Pstp > $SMX-2} {
					set SMX [expr 2*$SMX]
					.f$dbox2.c configure \
					 -scrollregion \
					 "[expr -$XSZ/2] 0 \
					  [expr $NPR*$XSZ] [expr $SMX*$YSZ]"
				}

				if { [info exists R($pstp,$pno)] == 0 } {
					if {$VERBOSE == 1} {
					#	set BoxFil $Yellow
					#	Michael Griffioen:
						if {[string first "~W " $stmnt] == 0} {
							set BoxFil $White
							set stmnt [string range $stmnt 3 end] 
						} else { if {[string first "~G " $stmnt] == 0} {
							set BoxFil $Green
							set stmnt [string range $stmnt 3 end]
						} else { if {[string first "~R " $stmnt] == 0} {
							set BoxFil $Red
							set stmnt [string range $stmnt 3 end]
						} else { if {[string first "~B " $stmnt] == 0} {
							set BoxFil $Blue
							set stmnt [string range $stmnt 3 end]
						} else { set BoxFil $Yellow } } } }
						set BoxLab $stmnt
					} else {
						set BoxLab $pstp
						set BoxFil $White
					}
					if {$SparseMsc == 0} {
						set HasBox($YLOC,[expr 1+$pno]) 1
					}
					set R($pstp,$pno) \
						[.f$dbox2.c create rectangle \
						 [expr $XLOC-20] $YLOC \
						 [expr $XLOC+20] \
						 [expr $YLOC+$YSZ] \
						 -outline $Blue -fill $BoxFil]
					set T($pstp,$pno) \
						[.f$dbox2.c create text \
						 $XLOC \
						 [expr $YLOC+$YSZ/2] \
						-font $SmallFont \
						-text $BoxLab]
					#if {$Pstp > $YNR-2} {
					#	.f$dbox2.c yview \
					#	 [expr ($Pstp-$YNR)]
					#}
				}
				if { $HAS == 3 } {
					.f$dbox2.c itemconfigure \
					 $R($pstp,$pno) \
					 -outline $Red -fill $Yellow
				}

				if {$SYMBOLIC} {
					.f$dbox2.c itemconfigure $T($pstp,$pno) \
						-font $SmallFont -text "$stmnt"
				} else {
				if {$VERBOSE == 0 } {
					.f$dbox2.c bind $T($pstp,$pno) <Any-Enter> "
					.f$dbox2.c itemconfigure $T($pstp,$pno) \
						-font $BigFont -text {$stmnt}
					.inp.t tag remove hilite 0.0 end
					if {[string first "pan_in" $Fnm] >= 0} {
						src_line {[format "%d" $pln]}
					}
					"
					.f$dbox2.c bind $T($pstp,$pno) <Any-Leave> "
					.f$dbox2.c itemconfigure $T($pstp,$pno) \
						-font $SmallFont -text {$pstp}
					"
				} else {
					.f$dbox2.c bind $T($pstp,$pno) <Any-Enter> "
					.inp.t tag remove hilite 0.0 end
					if {[string first "pan_in" $Fnm] >= 0} {
						src_line {[format "%d" $pln]}
					}
					"
				}
				}
			}

			set YLOC [expr $YLOC+$YSZ/2]
			if {$HAS == 1} {
				if { [info exists Q_add($inq)] == 0 } {
					set Q_add($inq) 0
					set Q_del($inq) 0
				}
				set Slot $Q_add($inq)
				incr Q_add($inq) 1

				set Mesg_y($inq,$Slot) $YLOC
				set Mesg_x($inq,$Slot) $XLOC
				set Q_val($inq,$Slot) $Mcont
			
				set Rem($inq,$Slot) \
					[.f$dbox2.c create text \
					[expr $XLOC-40] $YLOC \
					-font $SmallFont -text $stmnt]
			} elseif { $HAS == 2 } {
				if {$Easy} {
					set Slot $Q_del($inq)
					incr Q_del($inq) 1
				} else {
					for {set Slot $Q_del($inq)} \
						{$Slot < $Q_add($inq)} \
						{incr Slot} {
					if {$Q_val($inq,$Slot) == "_X_"} {
							incr Q_del($inq) 1
						} else {
							break
					}	}

					for {set Slot $Q_del($inq)} \
						{$Slot < $Q_add($inq)} \
						{incr Slot} {
					if {$Mcont == $Q_val($inq,$Slot)} {
						set Q_val($inq,$Slot) "_X_"
						break
					}	}
				}
				if {$Slot >= $Q_add($inq)} {
					add_log "<<error: cannot match $stmnt>>"
				} else {
					set TMP1 $Mesg_x($inq,$Slot)
					set TMP2 $Mesg_y($inq,$Slot)
					if {$XLOC < $TMP1} {
						set Delta -20
					} else {
						set Delta 20
					}
					.f$dbox2.c create line \
						[expr $TMP1+$Delta] $TMP2 \
						[expr $XLOC-$Delta] $YLOC \
						-fill $Red -width 2 \
						-arrow last -arrowshape {5 5 5}

					if {$SparseMsc == 0} {
						set TMP3 5
					} else {
						set TMP3 0
					}

					.f$dbox2.c coords $Rem($inq,$Slot) \
						[expr ($TMP1 + $XLOC)/2] \
						[expr ($TMP2 + $YLOC)/2 - $TMP3]
					.f$dbox2.c raise $Rem($inq,$Slot)
				}
			} }
		}
		if {$pln == 0 && ($gvars || $lvars)} {
			if {$Vvbox == 0} {
				if {$vv} { set Vvbox [mkbox "Variables" \
					"Vars" "var.out" 60 20 150 600] }
			} else {
				catch { .c$Vvbox.z.t delete 0.0 end }
			}
			if {$vv} {
			foreach el [lsort [array names Varnm]] {
				if {[string length $Varnm($el)] > 0} {
				catch { .c$Vvbox.z.t insert \
					end "$el = $Varnm($el)\n" }
			}	}
			foreach el [lsort [array names Queues]] {
				catch {
				.c$Vvbox.z.t insert end "queue $el ($Alias($el))\n"
				.c$Vvbox.z.t insert end "	$Queues($el)\n"
			}	}
			}
		}
	} else {
			set stepper 0
	}
		if {$isvar == 0} {
			if {$syntax == 1} {
				if {[string first "..." $line] < 0} {
					add_log "$line"
				catch { .c$sbox.z.t insert end "$line\n" }
				catch { .c$sbox.z.t yview -pickplace end }
				}
			} else {
				if {[string length $line] > 0} {
				catch { .c$sbox.z.t insert end "$line\n" }
				catch { .c$sbox.z.t yview -pickplace end }
				}
				if {$m_typ == 2 && \
					[string first "START OF CYCLE" $line] > 0} {
					catch { .c$dbox.z.t yview -pickplace end }
					catch { .c$dbox.z.t insert end "$line\n" }
					catch {
					set XLOC [expr $Seenpno*$XSZ+$XSZ/2]
					set YLOC [expr $Pstp*$YSZ+$YSZ/2]

					.f$dbox2.c create text \
						[expr $XLOC+$XSZ] $YLOC \
						-font $SmallFont \
						-text "Cycle/Waiting" \
						-fill $Red

					.f$dbox2.c create line \
						$XLOC $YLOC \
						[expr $XLOC+$XSZ/2] $YLOC \
						-fill $Red \
						-arrow first -arrowshape {5 5 5}
					}
					set HAS_CYCLE [expr $YLOC+1]
				}
				if {$m_typ == 2 && $HAS == 3 && $HAS_CYCLE != 0} {
					catch {
					set YLOC [expr $Pstp*$YSZ+$YSZ/2]
					set XLOC0 [expr $pno*$XSZ+$XSZ]
					set XLOC [expr $Seenpno*$XSZ+$XSZ]
					.f$dbox2.c create line \
						$XLOC0 [expr $YLOC-$YSZ/2] \
						$XLOC0 $YLOC \
						-fill $Red
					.f$dbox2.c create line \
						$XLOC0 $YLOC $XLOC $YLOC \
						-fill $Red

					set XLOC [expr $Seenpno*$XSZ+$XSZ]

					.f$dbox2.c create line \
						$XLOC $YLOC $XLOC \
						[expr $HAS_CYCLE-1] \
						-fill $Red
		}	}	}	}
		# mystery update:
		if {$tk_major >= 4 || $m_typ != 1} {
			update	;# tk 3.x can crash on this
		}

		if {$syntax == 0 \
		&&  $stop == 0 \
		&&  $stepped == 1 \
		&&  $stepper == 0 \
		&&  $dontwait == 0} {
			update	;# here it is harmless also with tk 3.x
			tkwait variable stepper
		}
	} else {
		if {$s_typ == 0 || $s_typ == 2} {
			add_log "<at end of run>"
		} else {
			add_log "<at end of trail>"
		}
		catch { addscales $dbox2 }
		if {$ebc} { barscales }
		update
		tkwait variable stepper
	}
	}
	# end of guided trail

	while {$stepper == 1} {
		tkwait variable stepper
	}
	set simruns 0
	catch {
		if { $m_typ == 1 } {
			foreach el [array names pbox] {
				catch { destroy .c$pbox($el) }
				catch { unset pbox($el) }
			}
		} elseif { $m_typ == 0 } {
			foreach el [array names Spbox] {
				catch { destroy .c$Spbox($el) }
				catch { unset Spbox($el) }
		}	}
	}
	catch { destroy .c$sbox }
	catch { destroy .c$dbox }
	catch { destroy .c$Vvbox }
	catch { unset .c$sbox }
	
	catch "close $fd"
	add_log "<done>"

	update
}

set PlaceMenu	"+150+150"

proc pickoption {nm} {
	global howmany Choice PlaceMenu

	catch {destroy .prompt}
	toplevel .prompt
	wm title .prompt "Select"
	wm iconname .prompt "Select"
	wm geometry .prompt $PlaceMenu

	text .prompt.t -relief raised -bd 2 \
		-width [string length $nm] -height 1 \
		-setgrid 1
	pack append .prompt .prompt.t { top expand fillx }
	.prompt.t insert end "$nm"
	for {set i 0} {$i <= $howmany} {incr i} {
		if {[info exists Choice($i)] \
		&&  $Choice($i) != 0 \
		&&  [string first "outside range" $Choice($i)] < 0 \
		&&  [string first "unexecutable," $Choice($i)] <= 0} {
			pack append .prompt \
			  [button .prompt.b$i -text "$i: $Choice($i)" \
			  -anchor w \
			  -command "set howmany $i" ] \
			  {top expand fillx}

			set j [string first "line " $Choice($i)]
			if {$j > 0} {
			  set k [string range $Choice($i) $j end]
			  scan $k "line %d" k
			  bind .prompt.b$i <Enter> "report $k"
			  bind .prompt.b$i <Leave> "report 0"
			  bind .prompt.b$i <ButtonRelease-1> "set howmany $i"
	}	}	}
	tkwait variable howmany
	set PlaceMenu [wm geometry .prompt]
	destroy .prompt
}

proc report {n} {
	.inp.t tag remove hilite 0.0 end
	if {$n > 0} { src_line $n }
}
# xspin
# validation options panel

set an_typ -1;	set cyc_typ 0
set ct_typ 0;	set et_typ 1
set st_typ 0;	set se_typ 0
set oct_typ -1;	# remembers last setting used for compilation
set nv_typ 1
set po_typ -1;	set vb_typ 0
set pr_typ 0;	set where 0
set vr_typ 0;	set xu_typ -1
set ur_typ 1;	set vbox 0
set killed 0;	set Job_Done 0;	set tcnt 0
set waitwhat "none"
set not_warned_yet 1

set e(0)	"Maximum Memory Usage (Mbytes), approx: "
set ival(0)	32

set e(1)	"Estimated State Space Size (states x 10^3): "
set ival(1)	500

set e(2)	"Maximum Search Depth (steps): "
set ival(2)	10000

set e(3)	"Extra Compile-Time Directives (Optional): "
set ival(3)	""

set e(4)	"Extra Run-Time Options (Optional): "
set ival(4)	""

set LastGenerate	""
set LastCompile	""
set NextCompile	""

proc syntax_check {} {
	global SPIN BG FG

	catch {destroy .h}
	toplevel .h

	wm title .h "Syntax"
	wm iconname .h "Syntax"
	text .h.t -setgrid 1 -width 50 -height 10 \
		 -background $FG -foreground $BG \
		-font -Adobe-Helvetica-Bold-R-Normal-*-120-* 
	mkpan_in
	add_log "$SPIN -a pan_in"
	catch "eval exec $SPIN -a pan_in >&pan.tmp"
	set cnt 0
	set maxln 50
	set ef [open pan.tmp r]
	.inp.t tag remove hilite 0.0 end
	.inp.t tag remove sel 0.0 end
	set pln 0
	while {[gets $ef line] > -1} {
		.h.t insert end "$line\n"
		if {[string first "line" $line] >= 0} {
			scan $line "spin: line %d" pln
			src_line [format "%d" $pln]
		}
		add_log "$line"
		if {[string length $line] > $maxln} {
			set maxln [string length $line]
			.h.t configure -width $maxln
		}
		incr cnt
	}
	close $ef

	if {$cnt == 0} {
		.h.t insert end "No syntax errors detected\n"
	}

	button .h.b -text "Ok" -command {destroy .h}
	pack append .h .h.t {top expand fill}
	pack append .h .h.b {top}

	add_log "syntax check completed"
	update
}

proc validation {} {
	global e ival hasfile
	global an_typ nv_typ
	global cyc_typ ct_typ
	global et_typ st_typ se_typ stop po_typ
	global vb_typ pr_typ vr_typ ur_typ xu_typ

	set stop 0
	set hasfile 0

	catch {destroy .v}
	toplevel .v

	wm title .v "Verification Options"
	wm iconname .v "VAL"
	wm geometry .v +150+150
#	wm minsize  .v 400 200

	frame .v.pan -relief flat

	# Initialize Sliders

	frame .v.memopts -relief flat;		# memory options panel
	frame .v.correct -relief flat;		# correctness properties
	frame .v.oframe  -relief raised -borderwidth 1m ;# error trail options
	frame .v.recomp  -relief raised -borderwidth 1m ;# recompilation
	frame .v.cframe  -relief raised -borderwidth 1m ;# compile time

	mkpan_in

	if {$an_typ == -1} {
		set an_typ 0
		set nv_typ [hasWord "never"]
		if {$nv_typ == 1} {
			if {[hasWord "accept.*:"]} {
				set an_typ 2
			}
		} elseif {[hasWord "progress.*:"]} {
				set an_typ 1
		}
	}
	if {$po_typ == -1} {
		if {[hasWord "_last"] \
		||  [hasWord "provided.*\\("] \
		||  [hasWord "enabled\\("]} {
			set po_typ 0
		} else {
			set po_typ 1
		}
	}
	if {$xu_typ == -1} {
		if {[hasWord "xr"] || [hasWord "xs"]} {
			set xu_typ 1
		} else {
			set xu_typ 0
		}
	}

	for {set x 0} {$x<=4} {incr x} {
		global e
		global ival

		frame .v.memopts.choice$x -relief flat
		entry .v.memopts.choice$x.e1 -relief sunken -width 20
		label .v.memopts.choice$x.e2 -text $e($x) -relief flat
		.v.memopts.choice$x.e1 insert end $ival($x)

		pack append .v.memopts.choice$x \
			.v.memopts.choice$x.e2 {left  frame w fillx} \
			.v.memopts.choice$x.e1 {right frame e fillx}

		pack append .v.memopts \
			.v.memopts.choice$x { top frame w pady 5 fillx}
	}
	set z .v.correct

	frame $z.rframe  -relief raised -borderwidth 1m

	label $z.rframe.lb \
		-font -Adobe-Helvetica-Bold-R-Normal-*-120-* \
		-text "Correctness Properties" \
		-relief sunken -borderwidth 1m
	radiobutton $z.rframe.sp -text "State Properties (Safety)" \
		-variable an_typ -value 0 \
		-relief flat \
		-command { set cyc_typ 0 }
	radiobutton $z.rframe.np -text "Non-Progress Cycles (Liveness)" \
		-variable an_typ -value 1 \
		-relief flat \
		-command { if {![hasWord "progress"] && $an_typ==1} then "complain4" }
	radiobutton $z.rframe.ac -text "Acceptance Cycles (Liveness)" \
		-variable an_typ -value 2 \
		-relief flat \
		-command { if {![hasWord "accept"] && $an_typ==2} then "complain1" }

	checkbutton $z.rframe.nv -text "Apply Never Claim (If Present)" \
		-variable nv_typ \
		-relief flat \
		-command { if {![hasWord "never"] && $nv_typ==1} then "complain2" }
	checkbutton $z.rframe.fc -text "Weak Fairness (for Liveness)" \
		-variable cyc_typ \
		-relief flat  \
		-command { if {$an_typ==0} then "set cyc_typ 0; complain3" }
	checkbutton $z.rframe.ur -text "Report Unreachable Code" \
		-variable ur_typ \
		-relief flat
	checkbutton $z.rframe.xu -text "Check xr/xs Assertions" \
		-variable xu_typ \
		-relief flat

	label $z.rframe.lb3 \
		-font -Adobe-Helvetica-Bold-R-Normal-*-120-* \
		-text "   Error Trapping   " \
		-relief sunken -borderwidth 1m
	radiobutton $z.rframe.c0 -text "Don't Stop at Errors" \
		-variable et_typ -value 0 \
		-relief flat 
	checkbutton $z.rframe.c0a -text "Save All Error-trails" \
		-variable st_typ \
		-relief flat 
	checkbutton $z.rframe.c0b -text "Find Shortest Trail (costly!)" \
		-variable se_typ \
		-relief flat 
	radiobutton $z.rframe.c1 -text "Stop at Error Nr:" \
		-variable et_typ -value 1 \
		-relief flat 
	entry $z.rframe.c2 -relief sunken -width 8
	$z.rframe.c2 insert end "1"

	pack append $z.rframe \
		$z.rframe.lb {top pady 4 frame w fillx} \
		$z.rframe.sp {top pady 4 frame w} \
		$z.rframe.np {top pady 4 frame w} \
		$z.rframe.ac {top pady 4 frame w} \
		$z.rframe.fc {top pady 4 frame w} \
		$z.rframe.nv {top pady 4 frame w} \
		$z.rframe.ur {top pady 4 frame w} \
		$z.rframe.xu {top pady 4 frame w filly}
	pack append $z.rframe \
		$z.rframe.lb3 { top expand fillx frame w} \
		$z.rframe.c0a {top pady 4 frame w} \
		$z.rframe.c0b {top pady 4 frame w} \
		$z.rframe.c0  {top pady 4 frame w} \
		$z.rframe.c2 {right} \
		$z.rframe.c1 {left expand}
	pack append $z $z.rframe {top frame nw filly}

	# Compile time options

	label .v.cframe.lb \
		-font -Adobe-Helvetica-Bold-R-Normal-*-120-* \
		-text "Search Mode" \
		-relief sunken -borderwidth 1m

	radiobutton .v.cframe.ea -text "Exhaustive" \
		-variable ct_typ -value 0 \
		-relief flat 
	radiobutton .v.cframe.sa -text "Supertrace/Bitstate" \
		-variable ct_typ -value 1 \
		-relief flat 

	pack append .v.cframe .v.cframe.lb {top pady 4 frame nw fillx} \
		.v.cframe.ea {top pady 4 frame nw} \
		.v.cframe.sa {top pady 4 frame nw}
#Type of Run:
	frame .v.pf -relief flat
	set z .v.pf
	frame $z.mesg -relief raised -borderwidth 1m;		# queue loss options
	frame $z.pframe -relief raised -borderwidth 1m
	label $z.pframe.lb2 \
		-font -Adobe-Helvetica-Bold-R-Normal-*-120-* \
		-text "   Type of Verification Run   " \
		-relief sunken -borderwidth 1m
	checkbutton $z.pframe.po -text "Reduced (Partial Orders)" \
		-variable po_typ \
		-relief flat 
	checkbutton $z.pframe.vb -text "Verbose (For Debugging)" \
		-variable vb_typ \
		-relief flat
	checkbutton $z.pframe.pr -text "Add Complexity Profiling" \
		-variable pr_typ \
		-relief flat
	checkbutton $z.pframe.vr -text "Compute Variable Ranges" \
		-variable vr_typ \
		-relief flat

	pack append $z.pframe \
		$z.pframe.lb2 {top fillx pady 4 frame w} \
		$z.pframe.po {top pady 4 frame w} \
		$z.pframe.vb  {top pady 4 frame w} \
		$z.pframe.pr  {top pady 4 frame w} \
		$z.pframe.vr  {top pady 4 frame w}
#Queues:
	set z .v.pf
	label $z.mesg.loss0 \
		-font -Adobe-Helvetica-Bold-R-Normal-*-120-* \
		-text "A Full Queue" \
		-relief sunken -borderwidth 1m
	radiobutton $z.mesg.loss1 -text "Blocks New Msgs" \
		-variable l_typ -value 0 \
		-relief flat
	radiobutton $z.mesg.loss2 -text "Loses New Msgs" \
		-variable l_typ -value 1 \
		-relief flat
	pack append $z.mesg \
		$z.mesg.loss0 {top pady 4 frame w fillx} \
		$z.mesg.loss1 {top pady 4 frame w} \
		$z.mesg.loss2 {top pady 4 frame w}
	pack append .v.pf \
		.v.pf.pframe  {top frame nw} \
		.v.pf.mesg {left frame nw expand fill}

#Memory:
	pack append .v \
		.v.memopts {top frame w}

#Correctness Properties:
	pack append .v \
		.v.correct {left frame nw}
#Search Mode:
	pack append .v \
		.v.cframe  {top frame w expand fill} \
		.v.pf { top frame w }

	pack append .v [button .v.run -text "Run" \
		-command "runval" ] {right frame se expand fill}
	pack append .v [button .v.exit -text "Cancel" \
		-command "stopval"] {right frame se}
	pack append .v [button .v.help -text "Help" -fg red \
		-command "roadmap2" ] {right frame se}

}

proc complain1 {} {
	add_log "warning: there are no accept labels"
}

proc complain2 {} {
	add_log "warning: there is no never claim"
}

proc complain3 {} {
	add_log "weak fairness is irrelevant to state properties"
}

proc complain4 {} {
	add_log "warning: there are no progress labels"
}

proc stopval {} {
	global stop
	global ival

	set ival(0) "[.v.memopts.choice0.e1 get]"
	set ival(1) "[.v.memopts.choice1.e1 get]"
	set ival(2) "[.v.memopts.choice2.e1 get]"
	set ival(3) "[.v.memopts.choice3.e1 get]"
	set ival(4) "[.v.memopts.choice4.e1 get]"

	set stop 1
	add_log "<stop verification>"
	catch {destroy .v}
}

proc log {n} {
	set m 1
	set cnt 0
	while {$m<$n} {
		set m [expr $m*2]
		incr cnt
	}
	return $cnt
}

proc bld_v_options {} {
	global an_typ;	global cyc_typ;
	global et_typ st_typ se_typ l_typ;
	global ct_typ;	global ival;
	global v_options a_options
	global c_options; global po_typ vb_typ pr_typ vr_typ ur_typ xu_typ;
	global ol_typ oct_typ nv_typ

	set ol_typ $l_typ
	set oct_typ $ct_typ

	set a_options "-a -X"
	if {$l_typ==1}	{ set a_options [format "%s -m" $a_options] }

	# the Mbytes scale resolution is in millions: 2^20

	set Mbytes [.v.memopts.choice0.e1 get]
	set Mbytes [expr 20 + log($Mbytes) / log(2) ]
		# 2^20 is 1 million
#	round down by removing the decimal part
	set foo [string first "." $Mbytes]
	if {$foo >= 0} {
		incr foo -1
		set Mbytes [string range $Mbytes 0 $foo]
	}
	unset foo

	# the Mstate scale resolution is in thousands: 2^10
	set Mstate [expr 10+[log [.v.memopts.choice1.e1 get]]]
	set Mdepth [.v.memopts.choice2.e1 get]
	set ival(0) "[.v.memopts.choice0.e1 get]"
	set ival(1) "[.v.memopts.choice1.e1 get]"
	set ival(2) "[.v.memopts.choice2.e1 get]"
	set ival(3) "[.v.memopts.choice3.e1 get]"
	set ival(4) "[.v.memopts.choice4.e1 get]"

	if {$ct_typ==1}	{
		set c_options [format "-DBITSTATE -DMEMCNT=%d" $Mbytes]

		# in supertrace mode: #bits ~~ 128x #states
		# (effectively the #bytes will be ~~ 16x #states)

		set Mstate [expr 7+$Mstate]
		set v_options "-X -m$Mdepth -w$Mstate"
	} else {
		set c_options [format "-DMEMCNT=%d" $Mbytes]

		# in exhaustive mode: #hashtable ~~ #states

		set v_options "-X -m$Mdepth -w$Mstate"
	}
	set c_options [format "-D_POSIX_SOURCE %s" $c_options]
	if {$an_typ==0}	{ set c_options [format "%s -DSAFETY" $c_options] }
	if {$an_typ==1}	{ set c_options [format "%s -DNP" $c_options] }
	if {$po_typ==0}	{ set c_options [format "%s -DNOREDUCE" $c_options] }
	if {$vb_typ==1}	{ set c_options [format "%s -DCHECK" $c_options] }
	if {$nv_typ==0}	{ set c_options [format "%s -DNOCLAIM" $c_options] }
	if {$se_typ!=0}	{ set c_options [format "%s -DREACH" $c_options] }
	if {$xu_typ==0 && $po_typ!=0}	{ set c_options [format "%s -DXUSAFE" $c_options] }
	if {$pr_typ==1}	{ set c_options [format "%s -DPEG" $c_options] }
	if {$vr_typ==1}	{ set c_options [format "%s -DVAR_RANGES" $c_options] }
	if {$cyc_typ==1}	{
		set c_options [format "%s -DNFAIR=3" $c_options]
	} else {
		set c_options [format "%s -DNOFAIR" $c_options]
	}
	set c_options [format "%s %s" $c_options [.v.memopts.choice3.e1 get]] 
	set v_options [format "%s %s" $v_options [.v.memopts.choice4.e1 get]] 

	if {$an_typ==1}	{ set v_options [format "%s -l" $v_options] }
	if {$an_typ==2}	{ set v_options [format "%s -a" $v_options] }
	if {$cyc_typ==1} { set v_options [format "%s -f" $v_options] }
	if {$ur_typ==0}	{ set v_options [format "%s -n" $v_options] }
	if {$st_typ==1}	{ set v_options [format "%s -e" $v_options] }
	if {$se_typ!=0}	{ set v_options [format "%s -i" $v_options] }
	if {$et_typ==0}	{ set v_options [format "%s -c0" $v_options] }
	if {$et_typ==1}	{
		set v_options [format "%s -c%s" $v_options \
			 [.v.correct.rframe.c2 get]]
	}
}

proc runval {} {
	global Unix GO32 CC SPIN
	global v_options a_options
	global c_options Job_Done
	global stop s_typ vbox waitwhat not_warned_yet
	global LastGenerate LastCompile NextCompile

	set stop 0
	bld_v_options
	catch {destroy .v}
	add_log "<starting verification>"

	set nochange [no_change]
	if {$nochange == 2} { return }	;# cannot proceed
	if {$nochange == 0} { set LastGenerate "" }

	if {$LastGenerate == $a_options} {
		add_log "<no code regeneration necessary>"
		set errmsg 0
	} else {
		set LastCompile ""
		add_log "$SPIN $a_options pan_in"; update
		if {$Unix} {
			catch "eval exec $SPIN $a_options pan_in" errmsg
		} else {
			catch "eval exec $SPIN $a_options pan_in >&pan.tmp"
			set errmsg [msg_file pan.tmp 0]
		}
		if {$stop} { update; return }
		if {[string length $errmsg]>0} {
			set foo [string first "Exit-Status 0" $errmsg]
			if {$foo<0} {
				add_log "$errmsg"
				update
				return
			}
			incr foo -2
			set errmsg [string range $errmsg 0 $foo]
			add_log "$errmsg"
	}	}
	set LastGenerate $a_options

	set NextCompile "$CC -o pan $c_options pan.c"

	if {$LastCompile == $NextCompile && [file exists pan]} {
		add_log "<no recompilation necessary>"
		set errmsg 0
	} else {
		add_log $NextCompile; update
		set nok 1
		if {$Unix} {
			set errmsg \
			[catch {eval exec $NextCompile &} cc_pid]
			if {[string compare $errmsg "0"] == 0} {
				set nok 0
			}
			set waitwhat "Compiling executable verifier"
		} else {
			catch "eval exec $NextCompile >&pan.tmp"
			set errmsg [msg_file pan.tmp 0]
			set cc_pid 0
			if {[string length $errmsg] == 0} {
				set nok 0
			}
			set waitwhat "Compiled executable verifier"
		}
		if {$stop || $nok} {
			add_log "$errmsg"
			update
			return
		}
		set Job_Done 0
		watch_job $cc_pid

		if {$Job_Done==0} {
			tkwait variable Job_Done
		}
		if {$Job_Done == -1} {
			add_log "aborted"
			return
		}
	}
	set LastCompile $NextCompile
	set NextCompile ""

	if {$Unix} {
		set PREFIX "time ./pan"
	} else {
		set PREFIX "$GO32 pan"
	}
	add_log "$PREFIX $v_options"; update
	set errmsg ""
	set stop 0
	update
	set errmsg [catch {eval exec $PREFIX $v_options >&pan.out &} pan_pid]
	if {$stop || $errmsg} {
		add_log "$errmsg"
		update
		return
	}

	set Job_Done 0
	set waitwhat "Running the verification"
#	incr pan_pid;	# `time' takes up 1 pid increment
	watch_job $pan_pid
	if {$Job_Done==0} {
		tkwait variable Job_Done
	}
	update
	set not_warned_yet 1
	set vbox [mkbox "Verification Output" "VerOut" "pan.out" 80 20]
	update
	after 1000 outcheck $vbox
	add_log "<done>"
	update
}

proc outcheck {vbox} {
	global stop where not_warned_yet

	set firstline 1
	set have_trail 0
	set po_red_viol 0
	if {$stop} return
	set fd [open "pan.out" r]
	while {[gets $fd line] > -1} {
		if {$firstline} {
			set firstline 0
			set nerr 0
			set trunc 0
			set nomem 0
			.inp.t tag remove hilite 0.0 end
			catch { .c$vbox.z.t delete 0.0 end }
		}
		catch { .c$vbox.z.t insert end "$line\n" }
		catch { .c$vbox.z.t yview -pickplace end }

		if {[string first "line" $line]>=0} {
			scan $line "\tline %d" where
			catch { src_line $where }
		}
		if {[string first "State-vector" $line] == 0} {
			set nerr [string first "errors:" $line]
			if {$nerr > 0} {
			set part [string range $line $nerr end]
			scan $part "errors: %d" nerr
			}
		}
		if {[string first "search depth too small" $line]>0} {
			set trunc 1
		}
		if {[string first "wrote pan_in.trail" $line]>0} {
			set have_trail 1
		}
		if {[string first "violated: access to chan" $line]>0} {
			set po_red_viol 1
		}
		if {[string first "out of memory" $line]>0} {
			set nomem 1
		}
		update
		if {$stop} return
	}
	if {$firstline} {
		catch "close $fd"
		if {$not_warned_yet} {
			add_log "waiting for output"
			add_log "\t(hit cancel in verification panel to abort)"
			set not_warned_yet 0
		}
		if {$stop == 0} { after 5000 outcheck $vbox }
	} else {
		catch { .c$vbox.z.t yview 0 }
		if {$have_trail && $po_red_viol==0 && ($nerr>0 || $trunc>0)} {
		catch { repeatbox $nerr $trunc $nomem }
		}
	}
}

proc src_line {s} {
	global FG BG
#puts "src $s"
	scan $s %d tgt_lnr

	if {$tgt_lnr > 0} {
	.inp.t tag add hilite $tgt_lnr.0 $tgt_lnr.end
	.inp.t tag configure hilite -background $FG -foreground $BG
	.inp.t yview -pickplace [expr $tgt_lnr-1]
	}
}

proc repeatbox { {nerr} {trunc} {nomem}} {
	global s_typ

	catch {destroy .rep}
	toplevel .rep

# puts "debug: repeatbox $nerr $trunc"

	wm title .rep "Suggested Action"
	wm iconname .rep "Suggest"
	wm geometry .rep +100+100

	button .rep.b0 -text "Dismiss" -command {destroy .rep}
	button .rep.b1 -text "Guided Simulation.." \
		-command {destroy .rep; simulation }
	button .rep.b2 -text "(Re)Verify.." \
		-command {destroy .rep; validation }

	if {$nerr>0} {
		message .rep.t -width 500 -text "\
Repeat the run with a different search
depth to find a shorter path to the error.

Or, perform a GUIDED simulation to retrace
the error found in this run."
		set s_typ 1

		pack append .rep .rep.t {top expand fill}
		pack append .rep .rep.b0 {right}
		pack append .rep .rep.b1 {right}
		pack append .rep .rep.b2 {right}
	} else {

		if {$nomem>0} {
			message .rep.t -width 500 -text "\
Increase the Maximum Memory Usage allotment, or
(if impossible) switch to Supertrace mode.

Then repeat the verification (with recompilation)."
		} else {
			message .rep.t -width 500 -text "\
Repeat the run with a different search
depth, to avoid truncation of the search."
		}
		pack append .rep .rep.t {top expand fill}
		pack append .rep .rep.b0 {right}
		pack append .rep .rep.b2 {right}
	}
}

proc watch_job {pid} {
	global killed tcnt waitwhat Unix Job_Done

	if {$Unix == 0} {
		add_log "$waitwhat"
		set Job_Done 1
		return
	}
	set w .cc_$pid
	toplevel $w
	wm title $w "Status Window"
	wm minsize $w 10 10
	wm geometry $w +90+90

	set killed 0
	message $w.t -aspect 1000 -relief raised -borderwidth 2

	button  $w.k -text "Abort" -command "killit $w $pid"
	pack append $w \
		$w.t {top} \
		$w.k {bot}
	set tcnt 0

	check_ps $w $pid
}

proc killit {w pid} {		;# only used on Unix systems
	global killed Job_Done
	catch { eval exec kill -2 $pid };
	$w.t configure -text "Killed Process $pid"
	add_log "Killed Process $pid"
	set killed $pid;	# cannot really kill cc jobs
	set Job_Done -1
	catch { destroy $w }
}

proc check_ps {w pid} {		;# only used on Unix systems
	global killed Job_Done tcnt waitwhat

	if {![winfo exists $w]} {
		set Job_Done -1
		return
	}
	if {$killed!=$pid} {
		catch {eval exec /bin/ps -e > tmp_pan_ps}
		set stat [catch {exec grep "^ *$pid " tmp_pan_ps} t]
		rmfile tmp_pan_ps
		if {$stat == 0} {
			incr tcnt 2
			$w.t configure \
			-text "$waitwhat $tcnt..."
			after 2000 check_ps $w $pid
			return
		} else {	;# process went away
			set Job_Done 1
	}	}
	catch { destroy $w }
}

# at the end - to make sure all
# proc declarations were seen

if {$argc == 1 && $Unix == 1} {
	set Fname "$argv"
	wm title . "SPIN CONTROL $xversion -- File: $Fname"
	cleanup
	cpfile $argv.trail pan_in.trail
	reopen
}

focus .inp.t
update
